# operators/_nonparametric.py
"""Classes for OpInf operators with no external parameter dependencies."""

__all__ = [
    "ConstantOperator",
    "LinearOperator",
    "QuadraticOperator",
    "CubicOperator",
    "QuarticOperator",
    "InputOperator",
    "StateInputOperator",
]

import itertools
import numpy as np
import scipy.linalg as la
import scipy.sparse as sparse
import scipy.special as special

from .. import utils
from ._base import OpInfOperator, InputMixin


# No dependence on state or input =============================================
class ConstantOperator(OpInfOperator):
    r"""Constant operator :math:`\Ophat_{\ell}(\qhat,\u) = \chat \in \RR^{r}`.

    Parameters
    ----------
    entries : (r,) ndarray or None
        Operator vector :math:`\chat`.

    Examples
    --------
    >>> import numpy as np
    >>> c = opinf.operators.ConstantOperator()
    >>> entries = np.random.random(10)          # Operator vector.
    >>> c.set_entries(np.random.random(10))
    >>> c.shape
    (10,)
    >>> out = c.apply()                         # "Apply" the operator.
    >>> np.allclose(out, entries)
    True
    """

    @staticmethod
    def _str(statestr=None, inputstr=None):
        return "c"

    @property
    def entries(self):
        r"""Operator vector :math:`\chat`."""
        return OpInfOperator.entries.fget(self)

    @entries.setter
    def entries(self, entries):
        """Set the ``entries`` attribute."""
        OpInfOperator.entries.fset(self, entries)

    @entries.deleter
    def entries(self):
        """Reset the ``entries`` attribute."""
        OpInfOperator.entries.fdel(self)

    @property
    def shape(self):
        r"""Shape :math:`(r,)` of the operator vector :math:`\chat`."""
        return OpInfOperator.shape.fget(self)

    def set_entries(self, entries):
        r"""Set the operator vector :math:`\chat`.

        Parameters
        ----------
        entries : (r,) ndarray
            Operator vector :math:`\chat`.
        """
        if sparse.issparse(entries):
            entries = entries.toarray()
        elif np.isscalar(entries):
            entries = np.atleast_1d(entries)
        self._validate_entries(entries)

        # Ensure that the operator is one-dimensional.
        if entries.ndim != 1:
            if entries.ndim == 2 and 1 in entries.shape:
                entries = np.ravel(entries)
            else:
                raise ValueError(
                    "ConstantOperator entries must be one-dimensional"
                )

        OpInfOperator.set_entries(self, entries)

    @utils.requires("entries")
    def apply(self, state=None, input_=None):
        r"""Apply the operator to the given state / input:
        :math:`\Ophat_{\ell}(\qhat,\u) = \chat`.

        Parameters
        ----------
        state : (r,) ndarray or None
            State vector.
        input_ : (m,) ndarray or None
            Input vector (not used).

        Returns
        -------
        out : (r,) ndarray
            :math:`\chat`.
        """
        if self.entries.shape[0] == 1:
            if state is None or np.isscalar(state):  # r = k = 1.
                return self.entries[0]
            return np.full_like(state, self.entries[0])  # r = 1, k > 1.
        # if state is None or np.ndim(state) == 1:
        #     return self.entries
        if np.ndim(state) == 2:  # r, k > 1.
            return np.outer(self.entries, np.ones(state.shape[-1]))
        return self.entries  # r > 1, k = 1.

    @utils.requires("entries")
    def galerkin(self, Vr, Wr=None):
        r"""Return the Galerkin projection of the operator,
        :math:`\chat = (\Wr\trp\Vr)^{-1}\Wr\trp\c`.

        Parameters
        ----------
        Vr : (n, r) ndarray
            Basis for the trial space.
        Wr : (n, r) ndarray or None
            Basis for the test space. If ``None``, defaults to ``Vr``.

        Returns
        -------
        projected : :class:`opinf.operators.ConstantOperator`
            Projected operator.
        """
        return self._galerkin(Vr, Wr, lambda c, V: c)

    @staticmethod
    def datablock(states, inputs=None):
        r"""Return the data matrix block corresponding to the operator,
        a row vector of ones.

        Since :math:`\Ophat_\ell(\qhat,\u) = \Ohat_{\ell}\d_{\ell}(\qhat,\u)`
        with :math:`\Ohat_{\ell} = \chat` and :math:`\d_{\ell}(\qhat,\u) = 1`,
        the data block is

        .. math::
           \D\trp
           = \left[\begin{array}{ccc}
           \d_{\ell}(\qhat_0,\u_0)
           & \cdots &
           \d_{\ell}(\qhat_{k-1},\u_{k-1})
           \end{array}\right]
           = \left[\begin{array}{ccc}
           1 & \cdots & 1
           \end{array}\right]
           \in \RR^{1 \times k}.

        Parameters
        ----------
        states : (r, k) or (k,) ndarray
            State vectors. Each column is a single state vector.
            If one dimensional, it is assumed that :math:`r = 1`.
        inputs : (m, k) or (k,) ndarray or None
            Input vectors (not used).

        Returns
        -------
        block : (1, k) ndarray
            Row vector of ones.
        """
        return np.ones((1, np.atleast_1d(states).shape[-1]))

    @staticmethod
    def operator_dimension(r=None, m=None):
        r"""Column dimension of the operator vector (always 1).

        Parameters
        ----------
        r : int
            State dimension.
        m : int or None
            Input dimension.
        """
        return 1


# Dependent on state but not on input =========================================
class LinearOperator(OpInfOperator):
    r"""Linear state operator :math:`\Ophat_{\ell}(\qhat,\u) = \Ahat\qhat`
    where :math:`\Ahat \in \RR^{r \times r}`.

    Parameters
    ----------
    entries : (r, r) ndarray or None
        Operator matrix :math:`\Ahat`.

    Examples
    --------
    >>> import numpy as np
    >>> A = opinf.operators.LinearOperator()
    >>> entries = np.random.random((10, 10))    # Operator matrix.
    >>> A.set_entries(entries)
    >>> A.shape
    (10, 10)
    >>> q = np.random.random(10)                # State vector.
    >>> out = A.apply(q)                        # Apply the operator to q.
    >>> np.allclose(out, entries @ q)
    True
    """

    @staticmethod
    def _str(statestr, inputstr=None):
        return f"A{statestr}"

    @property
    def entries(self):
        r"""Operator matrix :math:`\Ahat`."""
        return OpInfOperator.entries.fget(self)

    @entries.setter
    def entries(self, entries):
        """Set the ``entries`` attribute."""
        OpInfOperator.entries.fset(self, entries)

    @entries.deleter
    def entries(self):
        """Reset the ``entries`` attribute."""
        OpInfOperator.entries.fdel(self)

    @property
    def shape(self):
        r"""Shape :math:`(r, r)` of the operator matrix :math:`\Ahat`."""
        return OpInfOperator.shape.fget(self)

    def set_entries(self, entries):
        r"""Set the operator matrix :math:`\Ahat`.

        Parameters
        ----------
        entries : (r, r) ndarray
            Operator matrix :math:`\Ahat`.
        """
        if sparse.issparse(entries):
            if not isinstance(entries, sparse.csr_array):
                entries = entries.tocsr()
        elif np.isscalar(entries) or np.shape(entries) == (1,):
            entries = np.atleast_2d(entries)
        self._validate_entries(entries)

        # Ensure that the operator is two-dimensional and square.
        if entries.ndim != 2:
            raise ValueError("LinearOperator entries must be two-dimensional")
        if entries.shape[0] != entries.shape[1]:
            raise ValueError("LinearOperator entries must be square (r x r)")

        OpInfOperator.set_entries(self, entries)

    @utils.requires("entries")
    def apply(self, state, input_=None):
        r"""Apply the operator to the given state / input:
        :math:`\Ophat_{\ell}(\qhat,\u) = \Ahat\qhat`.

        Parameters
        ----------
        state : (r,) ndarray
            State vector.
        input_ : (m,) ndarray or None
            Input vector (not used).

        Returns
        -------
        out : (r,) ndarray
            Application :math:`\Ahat\qhat`.
        """
        if self.entries.shape[0] == 1:
            return self.entries[0, 0] * state  # r = 1.
        return self.entries @ state  # r > 1.

    @utils.requires("entries")
    def jacobian(self, state=None, input_=None):
        r"""Construct the state Jacobian of the operator:
        :math:`\ddqhat\Ophat_{\ell}(\qhat,\u)=\Ahat`.

        Parameters
        ----------
        state : (r,) ndarray or None
            State vector.
        input_ : (m,) ndarray or None
            Input vector (not used).

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian :math:`\Ahat`.
        """
        return self.entries

    @utils.requires("entries")
    def galerkin(self, Vr, Wr=None):
        r"""Return the Galerkin projection of the operator,
        :math:`\Ahat = (\Wr\trp\Vr)^{-1}\Wr\trp\A\Vr`.

        Parameters
        ----------
        Vr : (n, r) ndarray
            Basis for the trial space.
        Wr : (n, r) ndarray or None
            Basis for the test space. If ``None``, defaults to ``Vr``.

        Returns
        -------
        projected : :class:`opinf.operators.LinearOperator`
            Projected operator.
        """
        return self._galerkin(Vr, Wr, lambda A, V: A @ V)

    @staticmethod
    def datablock(states, inputs=None):
        r"""Return the data matrix block corresponding to the operator,
        the ``states``.

        Since :math:`\Ophat_\ell(\qhat,\u) = \Ohat_{\ell}\d_{\ell}(\qhat,\u)`
        with :math:`\Ohat_{\ell} = \Ahat` and
        :math:`\d_{\ell}(\qhat,\u) = \qhat`, the data block is

        .. math::
           \D\trp
           = \left[\begin{array}{ccc}
           \d_{\ell}(\qhat_0,\u_0)
           & \cdots &
           \d_{\ell}(\qhat_{k-1},\u_{k-1})
           \end{array}\right]
           = \left[\begin{array}{ccc}
           \qhat_0 & \cdots & \qhat_{k-1}
           \end{array}\right]
           \in \RR^{r \times k}.

        Parameters
        ----------
        states : (r, k) or (k,) ndarray
            State vectors. Each column is a single state vector.
            If one dimensional, it is assumed that :math:`r = 1`.
        inputs : (m, k) or (k,) ndarray or None
            Input vectors (not used).

        Returns
        -------
        state : (r, k) ndarray
            State vectors. Each column is a single state vector.
        """
        return np.atleast_2d(states)

    @staticmethod
    def operator_dimension(r, m=None):
        r"""Column dimension :math:`r` of the operator matrix :math:`\Ahat`.

        Parameters
        ----------
        r : int
            State dimension.
        m : int or None
            Input dimension.
        """
        return r


class QuadraticOperator(OpInfOperator):
    r"""Quadratic state operator
    :math:`\Ophat_{\ell}(\qhat,\u) = \Hhat[\qhat\otimes\qhat]`
    where :math:`\Hhat\in\RR^{r \times r^{2}}`.

    Internally, the action of the operator is computed as the product of an
    :math:`r \times r(r+1)/2` matrix :math:`\tilde{\H}` and a
    compressed version of the Kronecker product :math:`\qhat \otimes \qhat`.

    Parameters
    ----------
    entries : (r, r^2) or (r, r(r+1)/2) or (r, r, r) ndarray or None
        Operator matrix :math:`\Hhat`, its compressed representation
        :math:`\tilde{\H}`, or the equivalent symmetric tensor.

    Examples
    --------
    >>> import numpy as np
    >>> H = opinf.operators.QuadraticOperator()
    >>> entries = np.random.random((10, 100))   # Operator matrix.
    >>> H.set_entries(entries)
    >>> H.shape                                 # Compressed shape.
    (10, 55)
    >>> q = np.random.random(10)                # State vector.
    >>> out = H.apply(q)                        # Apply the operator to q.
    >>> np.allclose(out, entries @ np.kron(q, q))
    True
    """

    @staticmethod
    def _str(statestr, inputstr=None):
        return f"H[{statestr} ⊗ {statestr}]"

    def _clear(self):
        """Delete operator ``entries`` and related attributes."""
        self._mask = None
        self._prejac = None
        OpInfOperator._clear(self)

    def _precompute_jacobian_jit(self):
        """Compute (just in time) the pre-Jacobian tensor Jt such that
        Jt @ q = jacobian(q).
        """
        r = self.entries.shape[0]
        Ht = self.expand_entries(self.entries).reshape((r, r, r))
        self._prejac = Ht + Ht.transpose(0, 2, 1)

    @property
    def entries(self):
        r"""Internal representation :math:`\tilde{\H}` of the operator
        matrix :math:`\Hhat`.
        """
        return OpInfOperator.entries.fget(self)

    @entries.setter
    def entries(self, entries):
        """Set the ``entries`` attribute."""
        OpInfOperator.entries.fset(self, entries)

    @entries.deleter
    def entries(self):
        """Reset the ``entries`` attribute."""
        OpInfOperator.entries.fdel(self)

    @property
    def shape(self):
        r"""Shape :math:`(r, r(r+1)/2)` of the internal representation
        :math:`\tilde{\H}` of the operator matrix :math:`\Hhat`.
        """
        return OpInfOperator.shape.fget(self)

    def set_entries(self, entries):
        r"""Set the internal representation :math:`\tilde{\H}` of the operator
        matrix :math:`\Hhat`.

        Parameters
        ----------
        entries : (r, r^2) or (r, r(r+1)/2) or (r, r, r) ndarray
            Operator matrix :math:`\Hhat`, its compressed representation
            :math:`\tilde{\H}`, or the equivalent symmetric tensor.
        """
        if np.isscalar(entries) or np.shape(entries) == (1,):
            entries = np.atleast_2d(entries)
        self._validate_entries(entries)

        # Ensure that the operator has valid dimensions.
        if entries.ndim == 3 and len(set(entries.shape)) == 1:
            # Reshape (r x r x r) tensor.
            entries = entries.reshape((entries.shape[0], -1))
        if entries.ndim != 2:
            raise ValueError(
                "QuadraticOperator entries must be two-dimensional"
            )
        r, r2 = entries.shape
        if r2 == r**2:
            entries = self.compress_entries(entries)
        elif r2 != self.operator_dimension(r):
            raise ValueError("invalid QuadraticOperator entries dimensions")

        # Precompute compressed Kronecker product mask and Jacobian matrix.
        self._mask = self.ckron_indices(r)
        self._prejac = None

        OpInfOperator.set_entries(self, entries)

    @utils.requires("entries")
    def apply(self, state, input_=None):
        r"""Apply the operator to the given state / input:
        :math:`\Ophat_{\ell}(\qhat,\u) = \Hhat[\qhat\otimes\qhat]`

        Parameters
        ----------
        state : (r,) ndarray
            State vector.
        input_ : (m,) ndarray or None
            Input vector (not used).

        Returns
        -------
        out : (r,) ndarray
            Application :math:`\Hhat[\qhat\otimes\qhat]`.
        """
        if self.entries.shape[0] == 1:
            return self.entries[0, 0] * state**2  # r = 1
        return self.entries @ np.prod(state[self._mask], axis=1)

    @utils.requires("entries")
    def jacobian(self, state, input_=None):
        r"""Construct the state Jacobian of the operator:
        :math:`\ddqhat\Ophat_{\ell}(\qhat,\u)
        = \Hhat[(\I_r\otimes\qhat) + (\qhat\otimes\I_r)]`.

        Parameters
        ----------
        state : (r,) ndarray or None
            State vector.
        input_ : (m,) ndarray or None
            Input vector (not used).

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian
            :math:`\Hhat[(\I_r\otimes\qhat) + (\qhat\otimes\I_r)]`.
        """
        if self._prejac is None:
            self._precompute_jacobian_jit()
        return self._prejac @ np.atleast_1d(state)

    @utils.requires("entries")
    def galerkin(self, Vr, Wr=None):
        r"""Return the (Petrov-)Galerkin projection of the operator,
        :math:`\Hhat = (\Wr\trp\Vr)^{-1}\Wr\trp\H[\Vr\otimes\Vr]`.

        Parameters
        ----------
        Vr : (n, r) ndarray
            Basis for the trial space.
        Wr : (n, r) ndarray or None
            Basis for the test space. If ``None``, defaults to ``Vr``.

        Returns
        -------
        projected : :class:`opinf.operators.QuadraticOperator`
            Projected operator.
        """

        def _pg(H, V):
            return self.expand_entries(H) @ np.kron(V, V)

        return self._galerkin(Vr, Wr, _pg)

    @staticmethod
    def datablock(states, inputs=None):
        r"""Return the data matrix block corresponding to the operator,
        the Khatri--Rao product of the state with itself:
        :math:`\Qhat\odot\Qhat` where :math:`\Qhat` is ``states``.

        Since :math:`\Ophat_\ell(\qhat,\u) = \Ohat_{\ell}\d_{\ell}(\qhat,\u)`
        with :math:`\Ohat_{\ell} = \Hhat` and
        :math:`\d_{\ell}(\qhat,\u) = \qhat\otimes\qhat`,
        the data block should be

        .. math::
           \D\trp
           = \left[\begin{array}{ccc}
           \d_{\ell}(\qhat_0,\u_0)
           & \cdots &
           \d_{\ell}(\qhat_{k-1},\u_{k-1})
           \end{array}\right]
           = \left[\begin{array}{ccc}
           \qhat_0\otimes\qhat_0 & \cdots & \qhat_{k-1}\otimes\qhat_{k-1}
           \end{array}\right]
           \in\RR^{r^2 \times k}.

        Internally, a compressed Kronecker product :math:`\hat{\otimes}` with
        :math:`r(r+1)/2 < r^{2}` degrees of freedom is used for efficiency,
        hence the data block is actually

        .. math::
           \D\trp
           = \left[\begin{array}{ccc}
           \qhat_0\,\hat{\otimes}\,\qhat_0
           & \cdots &
           \qhat_{k-1}\,\hat{\otimes}\,\qhat_{k-1}
           \end{array}\right]
           \in\RR^{r(r+1)/2 \times k}.

        Parameters
        ----------
        states : (r, k) or (k,) ndarray
            State vectors. Each column is a single state vector.
            If one dimensional, it is assumed that :math:`r = 1`.
        inputs : (m, k) or (k,) ndarray or None
            Input vectors (not used).

        Returns
        -------
        product : (r(r+1)/2, k) ndarray
            Compressed Khatri--Rao product of ``states`` with itself.
        """
        return QuadraticOperator.ckron(np.atleast_2d(states))

    @staticmethod
    def operator_dimension(r, m=None):
        r"""Column dimension :math:`r(r+1)/2` of the internal representation
        :math:`\tilde{\H}` of the operator matrix :math:`\Hhat`.

        Parameters
        ----------
        r : int
            State dimension.
        m : int or None
            Input dimension.
        """
        return r * (r + 1) // 2

    # Utilities ---------------------------------------------------------------
    @staticmethod
    def ckron(state):
        r"""Calculate the compressed Kronecker product of a vector with itself.

        For a vector :math:`\qhat = [~\hat{q}_{1}~~\cdots~~\hat{q}_{r}~]\trp`,
        the Kronecker product of :math:`\qhat` with itself is given by

        .. math::
           \qhat \otimes \qhat
           = \left[\begin{array}{c}
               \hat{q}_{1}\qhat
               \\ \vdots \\
               \hat{q}_{r}\qhat
           \end{array}\right]
           =
           \left[\begin{array}{c}
               \hat{q}_{1}^{2} \\
               \hat{q}_{1}\hat{q}_{2} \\
               \vdots \\
               \hat{q}_{1}\hat{q}_{r} \\
               \hat{q}_{1}\hat{q}_{2} \\
               \hat{q}_{2}^{2} \\
               \vdots \\
               \hat{q}_{2}\hat{q}_{r} \\
               \vdots
               \hat{q}_{r}^{2}
           \end{array}\right] \in\RR^{r^2}.

        Cross terms :math:`\hat{q}_i \hat{q}_j` for :math:`i \neq j` appear
        twice in :math:`\qhat\otimes\qhat`.
        The *compressed Kronecker product* :math:`\qhat\,\hat{\otimes}\,\qhat`
        consists of the unique terms of :math:`\qhat\otimes\qhat`:

        .. math::
           \qhat\,\hat{\otimes}\,\qhat
           = \left[\begin{array}{c}
               \hat{q}_{1}^2
               \\
               \hat{q}_{2}\qhat_{1:2}
               \\ \vdots \\
               \hat{q}_{r}\qhat_{1:r}
           \end{array}\right]
           = \left[\begin{array}{c}
               \hat{q}_{1}^2 \\
               \hat{q}_{1}\hat{q}_{2} \\ \hat{q}_{2}^{2} \\
               \\ \vdots \\ \hline
               \hat{q}_{1}\hat{q}_{r} \\ \hat{q}_{2}\hat{q}_{r}
               \\ \vdots \\ \hat{q}_{r}^{2}
           \end{array}\right]
           \in \RR^{r(r+1)/2},
           \qquad
           \qhat_{1:i}
           = \left[\begin{array}{c}
               \hat{q}_{1} \\ \vdots \\ \hat{q}_{i}
           \end{array}\right]
           \in\RR^{i}.

        For matrices, the product is computed columnwise:

        .. math::
           \left[\begin{array}{c|c|c}
               & & \\
               \qhat_0 & \cdots & \qhat_{k-1}
               \\ & &
           \end{array}\right]
           \hat{\otimes}
           \left[\begin{array}{ccc}
               & & \\
               \qhat_0 & \cdots & \qhat_{k-1}
               \\ & &
           \end{array}\right]
           = \left[\begin{array}{ccc}
               & & \\
               \qhat_0\,\hat{\otimes}\,\qhat_0
               & \cdots &
               \qhat_{k-1}\,\hat{\otimes}\,\qhat_{k-1}
               \\ & &
           \end{array}\right]
           \in \RR^{r(r+1)/2 \times k}.

        Parameters
        ----------
        state : (r,) or (r, k) numpy.ndarray
            State vector or matrix where each column is a state vector.

        Returns
        -------
        product : (r(r+1)/2,) or (r(r+1)/2, k) ndarray
            The compressed Kronecker product of ``state`` with itself.
        """
        return np.concatenate(
            [state[i] * state[: i + 1] for i in range(state.shape[0])],
            axis=0,
        )

    @staticmethod
    def ckron_indices(r):
        """Construct a mask for efficiently computing the compressed Kronecker
        product.

        This method provides a faster way to evaluate :meth:`ckron`
        when the state dimension ``r`` is known *a priori*.

        Parameters
        ----------
        r : int
            State dimension.

        Returns
        -------
        mask : ndarray
            Compressed Kronecker product mask.

        Examples
        --------
        >>> from opinf.operators import QuadraticOperator
        >>> r = 20
        >>> mask = QuadraticOperator.ckron_indices(r)
        >>> q = np.random.random(r)
        >>> np.allclose(QuadraticOperator.ckron(q), np.prod(q[mask], axis=1))
        True
        """
        mask = np.zeros((r * (r + 1) // 2, 2), dtype=int)
        count = 0
        for i in range(r):
            for j in range(i + 1):
                mask[count, :] = (i, j)
                count += 1
        return mask

    @staticmethod
    def compress_entries(H):
        r"""Given :math:`\Hhat\in\RR^{a\times r^2}`, construct the matrix
        :math:`\tilde{\H}\in\RR^{a \times r(r+1)/2}` such that
        :math:`\Hhat[\qhat\otimes\qhat]
        = \tilde{\H}[\qhat\,\hat{\otimes}\,\qhat]`
        for all :math:`\qhat\in\RR^{r}` where :math:`\hat{\otimes}` is the
        compressed Kronecker product (see :meth:`ckron`).

        Parameters
        ----------
        H : (a, r^2) ndarray
            Matrix that acts on the full Kronecker product.

        Returns
        -------
        Hc : (a, r(r+1)/2) ndarray
            Matrix that acts on the compressed Kronecker product.

        Examples
        --------
        >>> from opinf.operators import QuadraticOperator
        >>> r = 20
        >>> H = np.random.random((r, r**2))
        >>> H.shape
        (20, 400)
        >>> Htilde = QuadraticOperator.compress_entries(H)
        >>> Htilde.shape
        (20, 210)
        >>> q = np.random.random(r)
        >>> Hq2 = H @ np.kron(q, q)
        >>> np.allclose(Hq2, Htilde @ QuadraticOperator.ckron(q))
        True
        """
        if np.ndim(H) == 1:
            H = np.atleast_2d(H)
        r2 = H.shape[1]
        if (r := int(round(r2 ** (1 / 2), 0))) ** 2 != r2:
            raise ValueError(
                f"invalid shape (a, r2) = {H.shape} "
                "with r2 not a perfect square"
            )
        Hc = np.empty((H.shape[0], r * (r + 1) // 2))

        fj = 0
        for i in range(r):
            for j in range(i + 1):
                if i == j:  # Place column for unique term.
                    Hc[:, fj] = H[:, (i * r) + j]
                else:  # Combine columns for repeated terms.
                    Hc[:, fj] = H[:, (i * r) + j] + H[:, (j * r) + i]
                fj += 1

        return Hc

    @staticmethod
    def expand_entries(Hc):
        r"""Given :math:`\tilde{\H}\in\RR^{a \times r(r+1)/2}`, construct the
        matrix :math:`\Hhat\in\RR^{a\times r^2}` such that
        :math:`\Hhat[\qhat\otimes\qhat]
        = \tilde{\H}[\qhat\,\hat{\otimes}\,\qhat]`
        for all :math:`\qhat\in\RR^{r}` where :math:`\hat{\otimes}` is the
        compressed Kronecker product (see :meth:`ckron`).

        Parameters
        ----------
        Hc : (a, r(r+1)/2) ndarray
            Matrix that acts on the compressed Kronecker product.

        Returns
        -------
        H : (a, r^2) ndarray
            Matrix that acts on the full Kronecker product.
            This matrix is "symmetric" in the sense that
            ``H.reshape((a, r, r))[i]`` is symmetric for `i = 0, ..., r`.

        Examples
        --------
        >>> from opinf.operators import QuadraticOperator
        >>> r = 20
        >>> Htilde = np.random.random((r, r * (r + 1) / 2))
        >>> Htilde.shape
        (20, 210)
        >>> H = QuadraticOperator.expand_entries(Htilde)
        >>> H.shape
        (20, 400)
        >>> q = np.random.random(r)
        >>> Hq2 = H @ np.kron(q, q)
        >>> np.allclose(Hq2, Htilde @ QuadraticOperator.ckron(q))
        True
        >>> np.all(QuadraticOperator.compress_entries(G) == Gtilde)
        True
        """
        if np.ndim(Hc) == 1:
            Hc = np.atleast_2d(Hc)
        b = Hc.shape[1]
        r = int(round(np.sqrt(1 + 8 * b) / 2 - 0.5, 0))
        if r * (r + 1) // 2 != b:
            raise ValueError(
                f"invalid shape (a, r2) = {Hc.shape} "
                "with r2 != r(r+1)/2 for any integer r"
            )

        H = np.empty((Hc.shape[0], r**2))
        fj = 0
        for i in range(r):
            for j in range(i + 1):
                if i == j:  # Place column for unique term.
                    H[:, (i * r) + j] = Hc[:, fj]
                else:  # Distribute columns equally for repeated terms.
                    fill = Hc[:, fj] / 2
                    H[:, (i * r) + j] = fill
                    H[:, (j * r) + i] = fill
                fj += 1

        return H


class CubicOperator(OpInfOperator):
    r"""Cubic state operator
    :math:`\Ophat_{\ell}(\qhat,\u) = \Ghat[\qhat\otimes\qhat\otimes\qhat]`
    where :math:`\Ghat\in\RR^{r \times r^{3}}`.

    Internally, the action of the operator is computed as the product of an
    :math:`r \times r(r+1)(r+2)/6` matrix :math:`\tilde{\G}` and a compressed
    version of the triple Kronecker product
    :math:`\qhat \otimes \qhat \otimes \qhat`.

    Parameters
    ----------
    entries : (r, r^3) or (r, r(r+1)(r+2)/6) or (r, r, r, r) ndarray or None
        Operator matrix :math:`\Ghat`, its compressed representation
        :math:`\tilde{\G}`, or the equivalent symmetric 4-tensor.

    Examples
    --------
    >>> import numpy as np
    >>> G = opinf.operators.CubicOperator()
    >>> entries = np.random.random((10, 1000))  # Operator matrix.
    >>> G.set_entries(entries)
    >>> G.shape                                 # Compressed shape.
    (10, 220)
    >>> q = np.random.random(10)                # State vector.
    >>> out = G.apply(q)                        # Apply the operator to q.
    >>> np.allclose(out, entries @ np.kron(q, np.kron(q, q)))
    True
    """

    @staticmethod
    def _str(statestr, inputstr=None):
        return f"G[{statestr} ⊗ {statestr} ⊗ {statestr}]"

    def _clear(self):
        """Delete operator ``entries`` and related attributes."""
        self._mask = None
        self._prejac = None
        OpInfOperator._clear(self)

    def _precompute_jacobian_jit(self):
        """Compute (just in time) the pre-Jacobian tensor Jt such that
        (Jt @ q) @ q = jacobian(q).
        """
        r = self.entries.shape[0]
        Gt = self.expand_entries(self.entries).reshape((r, r, r, r))
        self._prejac = Gt + Gt.transpose(0, 2, 1, 3) + Gt.transpose(0, 3, 1, 2)

    @property
    def entries(self):
        r"""Internal representation :math:`\tilde{\G}` of the operator
        matrix :math:`\Ghat`.
        """
        return OpInfOperator.entries.fget(self)

    @entries.setter
    def entries(self, entries):
        """Set the ``entries`` attribute."""
        OpInfOperator.entries.fset(self, entries)

    @entries.deleter
    def entries(self):
        """Reset the ``entries`` attribute."""
        OpInfOperator.entries.fdel(self)

    @property
    def shape(self):
        r"""Shape :math:`(r, r(r+1)(r+2)/6)` of the internal representation
        :math:`\tilde{\G}` of the operator matrix :math:`\Ghat`.
        """
        return OpInfOperator.shape.fget(self)

    def set_entries(self, entries):
        r"""Set the internal representation :math:`\tilde{\G}` of the operator
        matrix :math:`\Ghat`.

        Parameters
        ----------
        entries : (r, r^3) or (r, r(r+1)(r+2)/6) or (r, r, r, r) ndarray
            Operator matrix :math:`\Ghat`, its compressed representation
            :math:`\tilde{\G}`, or the equivalent symmetric 4-tensor.
        """
        if np.isscalar(entries) or np.shape(entries) == (1,):
            entries = np.atleast_2d(entries)
        self._validate_entries(entries)

        # Ensure that the operator has valid dimensions.
        if entries.ndim == 4 and len(set(entries.shape)) == 1:
            # Reshape (r x r x r x r) tensor.
            entries = entries.reshape((entries.shape[0], -1))
        if entries.ndim != 2:
            raise ValueError("CubicOperator entries must be two-dimensional")
        r, r3 = entries.shape
        if r3 == r**3:
            entries = self.compress_entries(entries)
        elif r3 != self.operator_dimension(r):
            raise ValueError("invalid CubicOperator entries dimensions")

        # Precompute compressed Kronecker product mask and Jacobian tensor.
        self._mask = self.ckron_indices(r)
        self._prejac = None

        OpInfOperator.set_entries(self, entries)

    @utils.requires("entries")
    def apply(self, state, input_=None):
        r"""Apply the operator to the given state / input:
        :math:`\Ophat_{\ell}(\qhat,\u) = \Ghat[\qhat\otimes\qhat\otimes\qhat]`.

        Parameters
        ----------
        state : (r,) ndarray
            State vector.
        input_ : (m,) ndarray or None
            Input vector (not used).

        Returns
        -------
        out : (r,) ndarray
            The evaluation :math:`\Ghat[\qhat\otimes\qhat\otimes\qhat]`.
        """
        if self.entries.shape[0] == 1:
            return self.entries[0, 0] * state**3  # r = 1.
        return self.entries @ np.prod(state[self._mask], axis=1)

    @utils.requires("entries")
    def jacobian(self, state, input_=None):
        r"""Construct the state Jacobian of the operator:
        :math:`\ddqhat\Ophat_{\ell}(\qhat,\u)
        = \Ghat[(\I_r\otimes\qhat\otimes\qhat)
        + (\qhat\otimes\I_r\otimes\qhat)
        + (\qhat\otimes\qhat\otimes\I_r)]`.

        Parameters
        ----------
        state : (r,) ndarray or None
            State vector.
        input_ : (m,) ndarray or None
            Input vector (not used).

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian
            :math:`\Ghat[(\I_r\otimes\qhat\otimes\qhat)
            + (\qhat\otimes\I_r\otimes\qhat)
            + (\qhat\otimes\qhat\otimes\I_r)]`.
        """
        if self._prejac is None:
            self._precompute_jacobian_jit()
        q_ = np.atleast_1d(state)
        return (self._prejac @ q_) @ q_

    @utils.requires("entries")
    def galerkin(self, Vr, Wr=None):
        r"""Return the Galerkin projection of the operator,
        :math:`\Ghat = (\Wr\trp\Vr)^{-1}\Wr\trp\G[\Vr\otimes\Vr\otimes\Vr]`.

        Parameters
        ----------
        Vr : (n, r) ndarray
            Basis for the trial space.
        Wr : (n, r) ndarray or None
            Basis for the test space. If ``None``, defaults to ``Vr``.

        Returns
        -------
        projected : :class:`opinf.operators.CubicOperator`
            Projected operator.
        """

        def _pg(G, V):
            return self.expand_entries(G) @ np.kron(V, np.kron(V, V))

        return self._galerkin(Vr, Wr, _pg)

    @staticmethod
    def datablock(states, inputs=None):
        r"""Return the data matrix block corresponding to the operator,
        the Khatri--Rao product of the state with itself three times:
        :math:`\Qhat\odot\Qhat\odot\Qhat` where :math:`\Qhat` is ``states``.

        Since :math:`\Ophat_\ell(\qhat,\u) = \Ohat_{\ell}\d_{\ell}(\qhat,\u)`
        with :math:`\Ohat_{\ell} = \Ghat` and
        :math:`\d_{\ell}(\qhat,\u) = \qhat\otimes\qhat\otimes\qhat`,
        the data block should be

        .. math::
           \D\trp
           = \left[\begin{array}{ccc}
           \d_{\ell}(\qhat_0,\u_0)
           & \cdots &
           \d_{\ell}(\qhat_{k-1},\u_{k-1})
           \end{array}\right]
           = \left[\begin{array}{ccc}
           \qhat_0\otimes\qhat_0\otimes\qhat_0
           & \cdots &
           \qhat_{k-1}\otimes\qhat_{k-1}\otimes\qhat_{k-1}
           \end{array}\right]
           \in \RR^{r^3 \times k}.

        Internally, a compressed triple Kronecker product with
        :math:`r(r+1)(r+2)/6 < r^{3}` degrees of freedom is used for
        efficiency, hence the data block is actually

        .. math::
           \D\trp
           = \left[\begin{array}{ccc}
           \qhat_0\,\hat{\otimes}\,\qhat_0\,\hat{\otimes}\,\qhat_0
           & \cdots &
           \qhat_{k-1}\,\hat{\otimes}\,\qhat_{k-1}\,\hat{\otimes}\,\qhat_{k-1}
           \end{array}\right]
           \in\RR^{r(r+1)(r+2)/6 \times k}.

        Parameters
        ----------
        states : (r, k) or (k,) ndarray
            State vectors. Each column is a single state vector.
            If one dimensional, it is assumed that :math:`r = 1`.
        inputs : (m, k) or (k,) ndarray or None
            Input vectors (not used).

        Returns
        -------
        product_ : (r(r+1)(r+2)/6, k) ndarray
            Compressed triple Khatri--Rao product of ``states`` with itself.
        """
        return CubicOperator.ckron(np.atleast_2d(states))

    @staticmethod
    def operator_dimension(r, m=None):
        r"""Column dimension :math:`r(r+1)(r+2)/6` of the internal
        representation :math:`\tilde{\G}` of the operator matrix :math:`\Ghat`.

        Parameters
        ----------
        r : int
            State dimension.
        m : int or None
            Input dimension.
        """
        return r * (r + 1) * (r + 2) // 6

    # Utilities ---------------------------------------------------------------
    @staticmethod
    def ckron(state):
        r"""Calculate the compressed cubic Kronecker product of a vector with
        itself.

        For a vector :math:`\qhat = [~\hat{q}_{1}~~\cdots~~\hat{q}_{r}~]\trp`,
        the cubic Kronecker product of :math:`\qhat` with itself is given by

        .. math::
           \qhat \otimes \qhat \otimes \qhat
           = \left[\begin{array}{c}
               \hat{q}_{1}(\qhat \otimes \qhat)
               \\ \vdots \\
               \hat{q}_{r}(\qhat \otimes \qhat)
           \end{array}\right]
           \in\RR^{r^3}.

        Cross terms :math:`\hat{q}_i \hat{q}_j \hat{q}_k` for :math:`i,j,k`
        not all equal appear multiple times in
        :math:`\qhat\otimes\qhat\otimes\qhat`.
        The *compressed cubic Kronecker product*
        :math:`\qhat\,\hat{\otimes}\,\qhat\,\hat{\otimes}\,\qhat`
        consists of the unique terms of :math:`\qhat\otimes\qhat\otimes\qhat`:

        .. math::
           \qhat\,\hat{\otimes}\,\qhat\,\hat{\otimes}\,\qhat
           = \left[\begin{array}{c}
               \hat{q}_{1}^3
               \\
               \hat{q}_{2}[\![\qhat\,\hat{\otimes}\,\qhat]\!]_{1:2}
               \\ \vdots \\
               \hat{q}_{r}[\![\qhat\,\hat{\otimes}\,\qhat]\!]_{1:r}
           \end{array}\right]
           \in \RR^{r(r+1)(r+2)/6}.

        See :meth:`opinf.operators.QuadraticOperator.ckron`.
        For matrices, the product is computed columnwise.

        Parameters
        ----------
        state : (r,) or (r, k) numpy.ndarray
            State vector or matrix where each column is a state vector.

        Returns
        -------
        product : (r(r+1)(r+2)/6,) or (r(r+1)(r+2)/6, k) ndarray
            The compressed triple Kronecker product of ``state`` with itself.
        """
        state2 = QuadraticOperator.ckron(state)
        lens = special.binom(np.arange(2, len(state) + 2), 2).astype(int)
        return np.concatenate(
            [state[i] * state2[: lens[i]] for i in range(state.shape[0])],
            axis=0,
        )

    @staticmethod
    def ckron_indices(r):
        """Construct a mask for efficiently computing the compressed Kronecker
        triple product.

        This method provides a faster way to evaluate :meth:`ckron`
        when the state dimension ``r`` is known *a priori*.

        Parameters
        ----------
        r : int
            State dimension.

        Returns
        -------
        mask : ndarray
            Compressed Kronecker product mask.

        Examples
        --------
        >>> from opinf.operators import CubicOperator
        >>> r = 20
        >>> mask = CubicOperator.kron_indices(r)
        >>> q = np.random.random(r)
        >>> np.allclose(CubicOperator.ckron(q), np.prod(q[mask], axis=1))
        True
        """
        mask = np.zeros((r * (r + 1) * (r + 2) // 6, 3), dtype=int)
        count = 0
        for i in range(r):
            for j in range(i + 1):
                for k in range(j + 1):
                    mask[count, :] = (i, j, k)
                    count += 1
        return mask

    @staticmethod
    def compress_entries(G):
        r"""Given :math:`\Ghat\in\RR^{a\times r^3}`, construct the matrix
        :math:`\tilde{\G}\in\RR^{a \times r(r+1)(r+2)/6}` such that
        :math:`\Ghat[\qhat\otimes\qhat\otimes\qhat]
        = \tilde{\G}[\qhat\,\hat{\otimes}\,\qhat\,\hat{\otimes}\,\qhat]`
        for all :math:`\qhat\in\RR^{r}`
        where :math:`\cdot\hat{\otimes}\cdot\hat{\otimes}\cdot` is the
        compressed cubic Kronecker product (see :meth:`ckron`).

        Parameters
        ----------
        G : (a, r^3) ndarray
            Matrix that acts on the full cubic Kronecker product.

        Returns
        -------
        Gc : (a, r(r+1)(r+2)/6) ndarray
            Matrix that acts on the compressed cubic Kronecker product.

        Examples
        --------
        >>> from opinf.operators import CubicOperator
        >>> r = 20
        >>> G = np.random.random((r, r**3))
        >>> G.shape
        (20, 8000)
        >>> Gtilde = CubicOperator.compress_entries(G)
        >>> Gtilde.shape
        (20, 1540)
        >>> q = np.random.random(r)
        >>> Gq3 = G @ np.kron(q, np.kron(q, q))
        >>> np.allclose(Gq3, Gtilde @ CubicOperator.ckron(q))
        True
        """
        if np.ndim(G) == 1:
            G = np.atleast_2d(G)
        r3 = G.shape[1]
        if (r := int(round(r3 ** (1 / 3), 0))) ** 3 != r3:
            raise ValueError(
                f"invalid shape (a, r3) = {G.shape} "
                "with r3 not a perfect cube"
            )
        Gc = np.empty((G.shape[0], r * (r + 1) * (r + 2) // 6))

        fj = 0
        for i in range(r):
            for j in range(i + 1):
                for k in range(j + 1):
                    idxs = set(itertools.permutations((i, j, k), 3))
                    Gc[:, fj] = np.sum(
                        [G[:, (a * r**2) + (b * r) + c] for a, b, c in idxs],
                        axis=0,
                    )
                    fj += 1

        return Gc

    @staticmethod
    def expand_entries(Gc):
        r"""Given :math:`\tilde{\G}\in\RR^{a \times r(r+1)(r+2)/6}`,
        construct the matrix :math:`\Ghat\in\RR^{a\times r^3}` such that
        :math:`\Ghat[\qhat\otimes\qhat\otimes\qhat]
        = \tilde{\G}[\qhat\,\hat{\otimes}\,\qhat\,\hat{\otimes}\,\qhat]`
        for all :math:`\qhat\in\RR^{r}`
        where :math:`\cdot\hat{\otimes}\cdot\hat{\otimes}\cdot` is the
        compressed cubic Kronecker product (see :meth:`ckron`).

        Parameters
        ----------
        Gc : (a, r(r+1)(r+2)/6) ndarray
            Matrix that acts on the compressed cubic Kronecker product.

        Returns
        -------
        G : (a, r^3) ndarray
            Matrix that acts on the full cubic Kronecker product.

        Examples
        --------
        >>> from opinf.operators import CubicOperator
        >>> r = 20
        >>> Gtilde = np.random.random((r, r * (r + 1) * (r + 2)/ 6))
        >>> Gtilde.shape
        (20, 1540)
        >>> G = CubicOperator.expand_entries(Gtilde)
        >>> G.shape
        (20, 8000)
        >>> q = np.random.random(r)
        >>> Gq3 = G @ np.kron(q, np.kron(q, q))
        >>> np.allclose(Gq3, Gtilde @ CubicOperator.ckron(q))
        True
        >>> np.all(CubicOperator.compress_entries(G) == Gtilde)
        True
        """
        if np.ndim(Gc) == 1:
            Gc = np.atleast_2d(Gc)
        b = Gc.shape[1]
        r = CubicOperator._rfromcompressed(b)
        if r * (r + 1) * (r + 2) // 6 != b:
            raise ValueError(
                f"invalid shape (a, r3) = {Gc.shape} "
                "with r3 != r(r+1)(r+2)/6 for any integer r"
            )

        G = np.empty((Gc.shape[0], r**3))
        fj = 0
        for i in range(r):
            for j in range(i + 1):
                for k in range(j + 1):
                    idxs = set(itertools.permutations((i, j, k), 3))
                    fill = Gc[:, fj] / len(idxs)
                    for a, b, c in idxs:
                        G[:, (a * r**2) + (b * r) + c] = fill
                    fj += 1

        return G

    @staticmethod
    def _rfromcompressed(b: int, maxiters: int = 10, tol: float = 0.25) -> int:
        """Compute r such that r(r+1)(r+2)/6 = b via 1D Newton's method."""
        r = int(b ** (1 / 3))
        _6b = 6 * b
        for _ in range(maxiters):
            _3r2 = 3 * r**2
            rnew = r - (r**3 + _3r2 + 2 * r - _6b) / (_3r2 + 6 * r + 2)
            if abs(r - rnew) < tol:
                return int(round(rnew, 0))
            r = rnew
        raise ValueError(  # pragma: no cover
            f"Newton solve for r such that r(r+1)(r+2)/6 = {b} failed"
        )


class QuarticOperator(OpInfOperator):
    r"""Quartic state operator
    :math:`\Ophat_{\ell}(\qhat,\u)
    = \Ghat[\qhat\otimes\qhat\otimes\qhat\otimes\qhat]`
    where :math:`\Ghat\in\RR^{r \times r^{4}}`.

    Internally, the action of the operator is computed as the product of an
    :math:`r \times r(r+1)(r+2)(r+3)/24` matrix :math:`\tilde{\G}` and a
    compressed version of the quadruple Kronecker product
    :math:`\qhat \otimes \qhat \otimes \qhat \otimes \qhat`.

    Parameters
    ----------
    entries : (r, r^4)/(r, r(r+1)(r+2)(r+2)/24)/(r, r, r, r, r) ndarray or None
        Operator matrix :math:`\Ghat`, its compressed representation
        :math:`\tilde{\G}`, or the equivalent symmetric 5-tensor.

    Examples
    --------
    >>> import numpy as np
    >>> G = opinf.operators.QuarticOperator()
    >>> entries = np.random.random((10, 10000))  # Operator matrix.
    >>> G.set_entries(entries)
    >>> G.shape                                  # Compressed shape.
    (10, 715)
    >>> q = np.random.random(10)                 # State vector.
    >>> out = G.apply(q)                         # Apply the operator to q.
    >>> np.allclose(out, entries @ np.kron(q, np.kron(q, np.kron(q, q))))
    True
    """

    @staticmethod
    def _str(statestr, inputstr=None):
        return f"G[{statestr} ⊗ {statestr} ⊗ {statestr} ⊗ {statestr}]"

    def _clear(self):
        """Delete operator ``entries`` and related attributes."""
        self._mask = None
        self._prejac = None
        OpInfOperator._clear(self)

    def _precompute_jacobian_jit(self):
        """Compute (just in time) the pre-Jacobian tensor Jt such that
        ((Jt @ q) @ q) @ q = jacobian(q).
        """
        r = self.entries.shape[0]
        Gt = self.expand_entries(self.entries).reshape((r, r, r, r, r))
        self._prejac = (
            Gt
            + Gt.transpose(0, 2, 1, 3, 4)
            + Gt.transpose(0, 3, 2, 1, 4)
            + Gt.transpose(0, 4, 2, 3, 1)
        )

    @property
    def entries(self):
        r"""Internal representation :math:`\tilde{\G}` of the operator
        matrix :math:`\Ghat`.
        """
        return OpInfOperator.entries.fget(self)

    @entries.setter
    def entries(self, entries):
        """Set the ``entries`` attribute."""
        OpInfOperator.entries.fset(self, entries)

    @entries.deleter
    def entries(self):
        """Reset the ``entries`` attribute."""
        OpInfOperator.entries.fdel(self)

    @property
    def shape(self):
        r"""Shape :math:`(r, r(r+1)(r+2)/6)` of the internal representation
        :math:`\tilde{\G}` of the operator matrix :math:`\Ghat`.
        """
        return OpInfOperator.shape.fget(self)

    def set_entries(self, entries):
        r"""Set the internal representation :math:`\tilde{\G}` of the operator
        matrix :math:`\Ghat`.

        Parameters
        ----------
        entries : (r, r^3) or (r, r(r+1)(r+2)/6) or (r, r, r, r) ndarray
            Operator matrix :math:`\Ghat`, its compressed representation
            :math:`\tilde{\G}`, or the equivalent symmetric 4-tensor.
        """
        if np.isscalar(entries) or np.shape(entries) == (1,):
            entries = np.atleast_2d(entries)
        self._validate_entries(entries)

        # Ensure that the operator has valid dimensions.
        if entries.ndim == 5 and len(set(entries.shape)) == 1:
            # Reshape (r x r x r x r x r) tensor.
            entries = entries.reshape((entries.shape[0], -1))
        if entries.ndim != 2:
            raise ValueError("QuarticOperator entries must be two-dimensional")
        r, r4 = entries.shape
        if r4 == r**4:
            entries = self.compress_entries(entries)
        elif r4 != self.operator_dimension(r):
            raise ValueError("invalid QuarticOperator entries dimensions")

        # Precompute compressed Kronecker product mask and Jacobian tensor.
        self._mask = self.ckron_indices(r)
        self._prejac = None

        OpInfOperator.set_entries(self, entries)

    @utils.requires("entries")
    def apply(self, state, input_=None):
        r"""Apply the operator to the given state / input:
        :math:`\Ophat_{\ell}(\qhat,\u) = \Ghat[\qhat\otimes\qhat\otimes\qhat]`.

        Parameters
        ----------
        state : (r,) ndarray
            State vector.
        input_ : (m,) ndarray or None
            Input vector (not used).

        Returns
        -------
        out : (r,) ndarray
            The evaluation :math:`\Ghat[\qhat\otimes\qhat\otimes\qhat]`.
        """
        if self.entries.shape[0] == 1:
            return self.entries[0, 0] * state**4  # r = 1.
        return self.entries @ np.prod(state[self._mask], axis=1)

    @utils.requires("entries")
    def jacobian(self, state, input_=None):
        r"""Construct the state Jacobian of the operator:
        :math:`\ddqhat\Ophat_{\ell}(\qhat,\u)
        = \Ghat[(\I_r\otimes\qhat\otimes\qhat)
        + (\qhat\otimes\I_r\otimes\qhat)
        + (\qhat\otimes\qhat\otimes\I_r)]`.

        Parameters
        ----------
        state : (r,) ndarray or None
            State vector.
        input_ : (m,) ndarray or None
            Input vector (not used).

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian
            :math:`\Ghat[(\I_r\otimes\qhat\otimes\qhat)
            + (\qhat\otimes\I_r\otimes\qhat)
            + (\qhat\otimes\qhat\otimes\I_r)]`.
        """
        if self._prejac is None:
            self._precompute_jacobian_jit()
        q = np.atleast_1d(state)
        return (self._prejac @ q) @ q @ q

    @utils.requires("entries")
    def galerkin(self, Vr, Wr=None):
        r"""Return the Galerkin projection of the operator,
        :math:`\Ghat
        = (\Wr\trp\Vr)^{-1}\Wr\trp\G[\Vr\otimes\Vr\otimes\Vr\otimes\Vr]`.

        Parameters
        ----------
        Vr : (n, r) ndarray
            Basis for the trial space.
        Wr : (n, r) ndarray or None
            Basis for the test space. If ``None``, defaults to ``Vr``.

        Returns
        -------
        projected : :class:`opinf.operators.CubicOperator`
            Projected operator.
        """

        def _pg(G, V):
            return self.expand_entries(G) @ np.kron(
                V, np.kron(V, np.kron(V, V))
            )

        return self._galerkin(Vr, Wr, _pg)

    @staticmethod
    def datablock(states, inputs=None):
        r"""Return the data matrix block corresponding to the operator,
        the Khatri--Rao product of the state with itself three times:
        :math:`\Qhat\odot\Qhat\odot\Qhat\odot\Qhat` where :math:`\Qhat` is
        ``states``.

        Since :math:`\Ophat_\ell(\qhat,\u) = \Ohat_{\ell}\d_{\ell}(\qhat,\u)`
        with :math:`\Ohat_{\ell} = \Ghat` and
        :math:`\d_{\ell}(\qhat,\u)
        = \qhat\otimes\qhat\otimes\qhat\otimes\qhat`,
        the data block should be

        .. math::
           \D\trp
           = \left[\begin{array}{ccc}
           \d_{\ell}(\qhat_0,\u_0)
           & \cdots &
           \d_{\ell}(\qhat_{k-1},\u_{k-1})
           \end{array}\right]
           = \left[\begin{array}{ccc}
           \qhat_0\otimes\qhat_0\otimes\qhat_0\otimes\qhat_0
           & \cdots &
           \qhat_{k-1}\otimes\qhat_{k-1}\otimes\qhat_{k-1}\otimes\qhat_{k-1}
           \end{array}\right]
           \in \RR^{r^4 \times k}.

        Internally, a compressed quadruple Kronecker product with
        :math:`r(r+1)(r+2)(r+3)/24 < r^{4}` degrees of freedom is used for
        efficiency, hence the data block is actually

        .. math::
           \D\trp
           = \left[\begin{array}{ccc}
           \qhat_0\,\hat{\otimes}\,
           \qhat_0\,\hat{\otimes}\,\qhat_0\,\hat{\otimes}\,\qhat_0
           & \cdots &
           \qhat_{k-1}\,\hat{\otimes}\,
           \qhat_{k-1}\,\hat{\otimes}\,\qhat_{k-1}\,\hat{\otimes}\,\qhat_{k-1}
           \end{array}\right]
           \in\RR^{r(r+1)(r+2)(r+3)/24 \times k}.

        Parameters
        ----------
        states : (r, k) or (k,) ndarray
            State vectors. Each column is a single state vector.
            If one dimensional, it is assumed that :math:`r = 1`.
        inputs : (m, k) or (k,) ndarray or None
            Input vectors (not used).

        Returns
        -------
        product_ : (r(r+1)(r+2)(r+3)/24, k) ndarray
            Compressed triple Khatri--Rao product of ``states`` with itself.
        """
        return QuarticOperator.ckron(np.atleast_2d(states))

    @staticmethod
    def operator_dimension(r, m=None):
        r"""Column dimension :math:`r(r+1)(r+2)(r+3)/24` of the internal
        representation :math:`\tilde{\G}` of the operator matrix :math:`\Ghat`.

        Parameters
        ----------
        r : int
            State dimension.
        m : int or None
            Input dimension.
        """
        return r * (r + 1) * (r + 2) * (r + 3) // 24

    # Utilities ---------------------------------------------------------------
    @staticmethod
    def ckron(state):
        r"""Calculate the compressed quartic Kronecker product of a vector with
        itself.

        For a vector :math:`\qhat = [~\hat{q}_{1}~~\cdots~~\hat{q}_{r}~]\trp`,
        the cubic Kronecker product of :math:`\qhat` with itself is given by

        .. math::
           \qhat \otimes \qhat \otimes \qhat
           = \left[\begin{array}{c}
               \hat{q}_{1}(\qhat \otimes \qhat \otimes \qhat)
               \\ \vdots \\
               \hat{q}_{r}(\qhat \otimes \qhat \otimes \qhat)
           \end{array}\right]
           \in\RR^{r^4}.

        Cross terms :math:`\hat{q}_i \hat{q}_j \hat{q}_k \hat{q}_\ell` for
        :math:`i,j,k,\ell`
        not all equal appear multiple times in
        :math:`\qhat\otimes\qhat\otimes\qhat\otimes\qhat`.
        The *compressed quartic Kronecker product*
        :math:`\qhat\,\hat{\otimes}\,\qhat\,\hat{\otimes}\,
        \qhat\,\hat{\otimes}\,\qhat`
        consists of the unique terms of
        :math:`\qhat\otimes\qhat\otimes\qhat\otimes\qhat`:

        .. math::
           \qhat\,\hat{\otimes}\,\qhat\,\hat{\otimes}\,\qhat
           = \left[\begin{array}{c}
               \hat{q}_{1}^4
               \\
               \hat{q}_{2}
               [\![\qhat\,\hat{\otimes}\,\qhat\,\hat{\otimes}\,\qhat]\!]_{1:2}
               \\ \vdots \\
               \hat{q}_{r}[\![\qhat\,\hat{\otimes}\,\qhat]\!]_{1:r}
           \end{array}\right]
           \in \RR^{r(r+1)(r+2)(r+3)/24}.

        See :meth:`opinf.operators.CubicOperator.ckron`.
        For matrices, the product is computed columnwise.

        Parameters
        ----------
        state : (r,) or (r, k) numpy.ndarray
            State vector or matrix where each column is a state vector.

        Returns
        -------
        product : (r(r+1)(r+2)(r+3)/24,) or (r(r+1)(r+2)(r+3)/24, k) ndarray
            The compressed triple Kronecker product of ``state`` with itself.
        """
        state3 = CubicOperator.ckron(state)
        lens = special.binom(np.arange(3, len(state) + 3), 3).astype(int)
        return np.concatenate(
            [state[i] * state3[: lens[i]] for i in range(state.shape[0])],
            axis=0,
        )

    @staticmethod
    def ckron_indices(r):
        """Construct a mask for efficiently computing the compressed Kronecker
        quadruple product.

        This method provides a faster way to evaluate :meth:`ckron`
        when the state dimension ``r`` is known *a priori*.

        Parameters
        ----------
        r : int
            State dimension.

        Returns
        -------
        mask : ndarray
            Compressed Kronecker product mask.

        Examples
        --------
        >>> from opinf.operators import QuarticOperator
        >>> r = 20
        >>> mask = QuarticOperator.kron_indices(r)
        >>> q = np.random.random(r)
        >>> np.allclose(QuarticOperator.ckron(q), np.prod(q[mask], axis=1))
        True
        """
        mask = np.zeros((r * (r + 1) * (r + 2) * (r + 3) // 24, 4), dtype=int)
        count = 0
        for i in range(r):
            for j in range(i + 1):
                for k in range(j + 1):
                    for ell in range(k + 1):
                        mask[count, :] = (i, j, k, ell)
                        count += 1
        return mask

    @staticmethod
    def compress_entries(G):
        r"""Given :math:`\Ghat\in\RR^{a\times r^4}`, construct the matrix
        :math:`\tilde{\G}\in\RR^{a \times r(r+1)(r+2)(r+3)/24}` such that
        :math:`\Ghat[\qhat\otimes\qhat\otimes\qhat\otimes\qhat]
        = \tilde{\G}[\qhat\,\hat{\otimes}\,
        \qhat\,\hat{\otimes}\,\qhat\,\hat{\otimes}\,\qhat]`
        for all :math:`\qhat\in\RR^{r}`
        where :math:`\cdot\hat{\otimes}\cdot\hat{\otimes}\cdot` is the
        compressed quartic Kronecker product (see :meth:`ckron`).

        Parameters
        ----------
        G : (a, r^4) ndarray
            Matrix that acts on the full quartic Kronecker product.

        Returns
        -------
        Gc : (a, r(r+1)(r+2)(r+3)/24) ndarray
            Matrix that acts on the compressed quartic Kronecker product.

        Examples
        --------
        >>> from opinf.operators import CubicOperator
        >>> r = 20
        >>> G = np.random.random((r, r**4))
        >>> G.shape
        (20, 160000)
        >>> Gtilde = CubicOperator.compress_entries(G)
        >>> Gtilde.shape
        (20, 8855)
        >>> q = np.random.random(r)
        >>> Gq3 = G @ np.kron(q, np.kron(q, q))
        >>> np.allclose(Gq3, Gtilde @ CubicOperator.ckron(q))
        True
        """
        if np.ndim(G) == 1:
            G = np.atleast_2d(G)
        r4 = G.shape[1]
        if (r := int(round(r4 ** (1 / 4), 0))) ** 4 != r4:
            raise ValueError(
                f"invalid shape (a, r4) = {G.shape} "
                "with r4 not a perfect quartic"
            )
        Gc = np.empty((G.shape[0], r * (r + 1) * (r + 2) * (r + 3) // 24))

        fj = 0
        for i in range(r):
            for j in range(i + 1):
                for k in range(j + 1):
                    for ell in range(k + 1):
                        idxs = set(itertools.permutations((i, j, k, ell), 4))
                        Gc[:, fj] = np.sum(
                            [
                                G[:, (a * r**3) + (b * r**2) + (c * r) + d]
                                for a, b, c, d in idxs
                            ],
                            axis=0,
                        )
                        fj += 1

        return Gc

    @staticmethod
    def expand_entries(Gc):
        r"""Given :math:`\tilde{\G}\in\RR^{a \times r(r+1)(r+2)(r+3)/24}`,
        construct the matrix :math:`\Ghat\in\RR^{a\times r^4}` such that
        :math:`\Ghat[\qhat\otimes\qhat\otimes\qhat\otimes\qhat]
        = \tilde{\G}[\qhat\,\hat{\otimes}\,\qhat\,\hat{\otimes}\,\qhat
        \,\hat{\otimes}\,\qhat]`
        for all :math:`\qhat\in\RR^{r}`
        where
        :math:`\cdot\hat{\otimes}\cdot\hat{\otimes}\cdot\hat{\otimes}\cdot` is
        the compressed quartic Kronecker product (see :meth:`ckron`).

        Parameters
        ----------
        Gc : (a, r(r+1)(r+2)(r+3)/24) ndarray
            Matrix that acts on the compressed cubic Kronecker product.

        Returns
        -------
        G : (a, r^4) ndarray
            Matrix that acts on the full cubic Kronecker product.

        Examples
        --------
        >>> from opinf.operators import QuarticOperator
        >>> r = 20
        >>> Gtilde = np.random.random((r, r*(r + 1)*(r + 2)*(r + 3)/24))
        >>> Gtilde.shape
        (20, 8855)
        >>> G = QuarticOperator.expand_entries(Gtilde)
        >>> G.shape
        (20, 160000)
        >>> q = np.random.random(r)
        >>> Gq4 = G @ np.kron(q, np.kron(q, np.kron(q, q)))
        >>> np.allclose(Gq4, Gtilde @ QuarticOperator.ckron(q))
        True
        >>> np.all(QuarticOperator.compress_entries(G) == Gtilde)
        True
        """
        if np.ndim(Gc) == 1:
            Gc = np.atleast_2d(Gc)
        b = Gc.shape[1]
        r = QuarticOperator._rfromcompressed(b)
        if r * (r + 1) * (r + 2) * (r + 3) // 24 != b:
            raise ValueError(
                f"invalid shape (a, r4) = {Gc.shape} "
                "with r4 != r(r+1)(r+2)(r+3)/24 for any integer r"
            )

        G = np.empty((Gc.shape[0], r**4))
        fj = 0
        for i in range(r):
            for j in range(i + 1):
                for k in range(j + 1):
                    for ell in range(k + 1):
                        idxs = set(itertools.permutations((i, j, k, ell), 4))
                        fill = Gc[:, fj] / len(idxs)
                        for a, b, c, d in idxs:
                            G[:, (a * r**3) + (b * r**2) + (c * r) + d] = fill
                        fj += 1

        return G

    @staticmethod
    def _rfromcompressed(b: int, maxiters: int = 10, tol: float = 0.25) -> int:
        """Compute r such that r(r+1)(r+2)(r+3)/24 = b
        via 1D Newton's method.
        """
        r = int(b ** (1 / 4))
        _24b = 24 * b
        for _ in range(maxiters):
            rnew = r - (r**4 + 6 * r**3 + 11 * r**2 + 6 * r - _24b) / (
                4 * r**3 + 18 * r**2 + 22 * r + 6
            )
            if abs(r - rnew) < tol:
                return int(round(rnew, 0))
            r = rnew
        raise ValueError(  # pragma: no cover
            f"Newton solve for r such that r(r+1)(r+2)(r+3)/24 = {b} failed"
        )


# Dependent on input but not on state =========================================
class InputOperator(OpInfOperator, InputMixin):
    r"""Linear input operator :math:`\Ophat_{\ell}(\qhat,\u) = \Bhat\u`
    where :math:`\Bhat \in \RR^{r \times m}`.

    Parameters
    ----------
    entries : (r, m) ndarray or None
        Operator matrix :math:`\Bhat`.

    Examples
    --------
    >>> import numpy as np
    >>> B = opinf.operators.LinearOperator()
    >>> entries = np.random.random((10, 3))     # Operator matrix.
    >>> B.set_entries(entries)
    >>> B.shape
    (10, 3)
    >>> u = np.random.random(3)                 # Input vector.
    >>> out = B.apply(None, u)                  # Apply the operator to u.
    >>> np.allclose(out, entries @ u)
    True
    """

    @property
    def input_dimension(self):
        r"""Dimension :math:`m` of the input :math:`\u` that the operator
        acts on.
        """
        return None if self.entries is None else self.entries.shape[1]

    @staticmethod
    def _str(statestr, inputstr):
        return f"B{inputstr}"

    @property
    def entries(self):
        r"""Operator matrix :math:`\Bhat`."""
        return OpInfOperator.entries.fget(self)

    @entries.setter
    def entries(self, entries):
        """Set the ``entries`` attribute."""
        OpInfOperator.entries.fset(self, entries)

    @entries.deleter
    def entries(self):
        """Reset the ``entries`` attribute."""
        OpInfOperator.entries.fdel(self)

    @property
    def shape(self):
        r"""Shape :math:`(r, m)` of the operator matrix :math:`\Bhat`."""
        return OpInfOperator.shape.fget(self)

    def set_entries(self, entries):
        r"""Set the operator matrix :math:`\Bhat`.

        Parameters
        ----------
        entries : (r, m) ndarray
            Operator matrix :math:`\Bhat`.
        """
        if np.isscalar(entries) or np.shape(entries) == (1,):
            entries = np.atleast_2d(entries)
        self._validate_entries(entries)

        # Ensure that the operator is two-dimensional.
        if entries.ndim == 1:
            # Assumes r = entries.size, m = 1.
            entries = entries.reshape((-1, 1))
        if entries.ndim != 2:
            raise ValueError("InputOperator entries must be two-dimensional")

        OpInfOperator.set_entries(self, entries)

    @utils.requires("entries")
    def apply(self, state, input_):
        r"""Apply the operator to the given state / input:
        :math:`\Ophat_{\ell}(\qhat,\u) = \Bhat\u`.

        Parameters
        ----------
        state : (r,) ndarray
            State vector (not used).
        input_ : (m,) ndarray
            Input vector.

        Returns
        -------
        out : (r,) ndarray
            Application :math:`\Bhat\u`.
        """
        if self.entries.shape[1] == 1 and (dim := np.ndim(input_)) != 2:
            if self.entries.shape[0] == 1:
                return self.entries[0, 0] * input_  # r = m = 1.
            if dim == 1 and input_.size > 1:  # r, k > 1, m = 1.
                return np.outer(self.entries[:, 0], input_)
            return self.entries[:, 0] * input_  # r > 1, m = k = 1.
        return self.entries @ input_  # m > 1.

    @utils.requires("entries")
    def galerkin(self, Vr, Wr=None):
        r"""Return the Galerkin projection of the operator,
        :math:`\Bhat = (\Wr\trp\Vr)^{-1}\Wr\trp\B`.

        Parameters
        ----------
        Vr : (n, r) ndarray
            Basis for the trial space.
        Wr : (n, r) ndarray or None
            Basis for the test space. If ``None``, defaults to ``Vr``.

        Returns
        -------
        projected : :class:`opinf.operators.InputOperator`
            Projected operator.
        """
        return self._galerkin(Vr, Wr, lambda B, V: B)

    @staticmethod
    def datablock(states, inputs):
        r"""Return the data matrix block corresponding to the operator,
        the ``inputs``.

        Since :math:`\Ophat_\ell(\qhat,\u) = \Ohat_{\ell}\d_{\ell}(\qhat,\u)`
        with :math:`\Ohat_{\ell} = \Bhat` and
        :math:`\d_{\ell}(\qhat,\u) = \u`, the data block is

        .. math::
           \D\trp
           = \left[\begin{array}{ccc}
           \d_{\ell}(\qhat_0,\u_0)
           & \cdots &
           \d_{\ell}(\qhat_{k-1},\u_{k-1})
           \end{array}\right]
           = \left[\begin{array}{ccc}
           \u_0 & \cdots & \u_{k-1}
           \end{array}\right]
           \in \RR^{r \times k}.

        Parameters
        ----------
        states : (r, k) or (k,) ndarray
            State vectors (not used).
        inputs : (m, k) or (k,) ndarray
            Input vectors. Each column is a single input vector.
            If one dimensional, it is assumed that :math:`m = 1`.

        Returns
        -------
        inputs : (m, k) ndarray
            Input vectors. Each column is a single input vector.
        """
        return np.atleast_2d(inputs)

    @staticmethod
    def operator_dimension(r, m):
        r"""Column dimension :math:`m` of the operator matrix :math:`\Bhat`.

        Parameters
        ----------
        r : int
            State dimension.
        m : int or None
            Input dimension.
        """
        return m


# Dependent on both state and input ===========================================
class StateInputOperator(OpInfOperator, InputMixin):
    r"""Linear state / input interaction operator
    :math:`\Ophat_{\ell}(\qhat,\u) = \Nhat[\u\otimes\qhat]`
    where :math:`\Nhat \in \RR^{r \times rm}`.

    Parameters
    ----------
    entries : (r, rm) ndarray or None
        Operator matrix :math:`\Nhat`.

    Examples
    --------
    >>> import numpy as np
    >>> N = opinf.operators.StateInputOperator()
    >>> entries = np.random.random((10, 3))
    >>> N.set_entries(entries)
    >>> N.shape
    (10, 3)
    >>> q = np.random.random(10)                # State vector.
    >>> u = np.random.random(3)                 # Input vector.
    >>> out = N.apply(q, u)                     # Apply the operator to (q,u).
    >>> np.allclose(out, entries @ np.kron(u, q))
    True
    """

    @property
    def input_dimension(self):
        r"""Dimension :math:`m` of the input :math:`\u` that the operator
        acts on.
        """
        if self.entries is None:
            return None
        return self.entries.shape[1] // self.entries.shape[0]

    @staticmethod
    def _str(statestr, inputstr):
        return f"N[{inputstr} ⊗ {statestr}]"

    @property
    def entries(self):
        r"""Operator matrix :math:`\Nhat`."""
        return OpInfOperator.entries.fget(self)

    @entries.setter
    def entries(self, entries):
        """Set the ``entries`` attribute."""
        OpInfOperator.entries.fset(self, entries)

    @entries.deleter
    def entries(self):
        """Reset the ``entries`` attribute."""
        OpInfOperator.entries.fdel(self)

    @property
    def shape(self):
        r"""Shape :math:`(r, rm)` of the operator matrix :math:`\Nhat`."""
        return OpInfOperator.shape.fget(self)

    def set_entries(self, entries):
        r"""Set the operator matrix :math:`\Nhat`.

        Parameters
        ----------
        entries : (r, rm) ndarray
            Operator matrix :math:`\Nhat`.
        """
        if np.isscalar(entries) or np.shape(entries) == (1,):
            entries = np.atleast_2d(entries)
        self._validate_entries(entries)

        # Ensure that the operator has valid dimensions.
        if entries.ndim != 2:
            raise ValueError(
                "StateInputOperator entries must be two-dimensional"
            )
        r, rm = entries.shape
        m = rm // r
        if rm != r * m:
            raise ValueError("invalid StateInputOperator entries dimensions")

        OpInfOperator.set_entries(self, entries)

    @utils.requires("entries")
    def apply(self, state, input_):
        r"""Apply the operator to the given state / input:
        :math:`\Ophat_{\ell}(\qhat,\u) = \Nhat[\u\otimes\qhat]`.

        Parameters
        ----------
        state : (r,) ndarray
            State vector.
        input_ : (m,) ndarray
            Input vector.

        Returns
        -------
        out : (r,) ndarray
            The evaluation :math:`\Nhat[\u\otimes\qhat]`.
        """
        # Determine if arguments represent one snapshot or several.
        multi = (sdim := np.ndim(state)) > 1
        multi |= (idim := np.ndim(input_)) > 1
        multi |= self.shape[0] == 1 and sdim == 1 and state.shape[0] > 1
        multi |= self.shape[1] == 1 and idim == 1 and input_.shape[0] > 1
        single = not multi

        if self.shape[1] == 1:
            return self.entries[0, 0] * input_ * state  # r = m = 1.
        if single:
            return self.entries @ np.kron(input_, state)  # k = 1, rm > 1.
        Q_ = np.atleast_2d(state)
        U = np.atleast_2d(input_)
        return self.entries @ la.khatri_rao(U, Q_)  # k > 1, rm > 1.

    @utils.requires("entries")
    def jacobian(self, state, input_):
        r"""Construct the state Jacobian of the operator:
        :math:`\ddqhat\Ophat_{\ell}(\qhat,\u) = \sum_{i=1}^{m}u_{i}\Nhat_{i}`
        where :math:`\Nhat=[~\Nhat_{1}~~\cdots~~\Nhat_{m}~]`
        and each :math:`\Nhat_i\in\RR^{r\times r},~i=1,\ldots,m`.

        Parameters
        ----------
        state : (r,) ndarray or None
            State vector.
        input_ : (m,) ndarray or None
            Input vector (not used).

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian :math:`\sum_{i=1}^{m}u_{i}\Nhat_{i}`.
        """
        r, rm = self.entries.shape
        m = rm // r
        u = np.atleast_1d(input_)
        if u.shape[0] != m:
            raise ValueError("invalid input_ shape")
        return np.sum(
            [um * Nm for um, Nm in zip(u, np.split(self.entries, m, axis=1))],
            axis=0,
        )

    @utils.requires("entries")
    def galerkin(self, Vr, Wr=None):
        r"""Return the Galerkin projection of the operator,
        :math:`\Nhat = (\Wr\trp\Vr)^{-1}\Wr\trp\N[\I_{m}\otimes\Vr]`.

        Parameters
        ----------
        Vr : (n, r) ndarray
            Basis for the trial space.
        Wr : (n, r) ndarray or None
            Basis for the test space. If ``None``, defaults to ``Vr``.

        Returns
        -------
        projected : :class:`opinf.operators.StateInputOperator`
            Projected operator.
        """

        def _pg(N, V):
            r, rm = N.shape
            m = rm // r
            return N @ np.kron(np.eye(m), V)

        return self._galerkin(Vr, Wr, _pg)

    @staticmethod
    def datablock(states, inputs):
        r"""Return the data matrix block corresponding to the operator,
        the Khatri--Rao product :math:`\U\odot\Qhat` where
        :math:`\Qhat` is ``states`` and :math:`\U` is ``inputs``.

        Since :math:`\Ophat_\ell(\qhat,\u) = \Ohat_{\ell}\d_{\ell}(\qhat,\u)`
        with :math:`\Ohat_{\ell} = \Nhat` and
        :math:`\d_{\ell}(\qhat,\u) = \u\otimes\qhat`, the data block is

        .. math::
           \D\trp
           = \left[\begin{array}{ccc}
           \d_{\ell}(\qhat_0,\u_0)
           & \cdots &
           \d_{\ell}(\qhat_{k-1},\u_{k-1})
           \end{array}\right]
           = \left[\begin{array}{ccc}
           \u_0 \otimes \qhat_0 & \cdots & \u_{k-1} \otimes \qhat_{k-1}
           \end{array}\right]
           \in \RR^{rm \times k}.

        Parameters
        ----------
        states : (r, k) or (k,) ndarray
            State vectors (not used).
            If one dimensional, it is assumed that :math:`r = 1`.
        inputs : (m, k) or (k,) ndarray or None
            Input vectors. Each column is a single input vector.
            If one dimensional, it is assumed that :math:`m = 1`.

        Returns
        -------
        product_ : (m, k) ndarray or None
            Compressed Khatri-Rao product of the ``input_`` and the ``states``.
        """
        return la.khatri_rao(np.atleast_2d(inputs), np.atleast_2d(states))

    @staticmethod
    def operator_dimension(r, m):
        r"""Column dimension :math:`rm` of the operator matrix :math:`\Nhat`.

        Parameters
        ----------
        r : int
            State dimension.
        m : int or None
            Input dimension.
        """
        return r * m



# ddt/_base.py
"""Template for time derivative estimators."""

__all__ = [
    "DerivativeEstimatorTemplate",
]


import abc
import collections
import numpy as np
import scipy.linalg as la
import matplotlib.pyplot as plt

from .. import errors, utils


class DerivativeEstimatorTemplate(abc.ABC):
    r"""Template for time derivative estimators.

    Operator Inference for time-continuous (semi-discrete) models requires
    state snapshots and their time derivatives in order to learn operator
    entries via regression. This class is a template for estimating the first
    time derivative of state snapshots. Specifically, from a collection of
    snapshots :math:`\qhat_0,\ldots,\qhat_{k-1}\in\RR^r` representing the state
    at time instances :math:`t_0,\ldots,t_{k-1}`, the goal is to estimate

    .. math::
        \dot{\qhat}_j \approx \ddt\qhat(t)\bigg|_{t = t_j} \in \RR^{r}

    for :math:`j = 0, \ldots, k - 1`.

    Depending on the estimation strategy, the derivatives may only be computed
    for a subset of the states. For example, a first-order backward difference
    may omit an estimate for :math:`\dot{\qhat}_0`.

    Parameters
    ----------
    time_domain : (k,) ndarray
        Time domain of the snapshot data.
    """

    __tests = (
        (
            r"$f(t) = t$",
            lambda t: t,
            np.ones_like,
        ),
        (
            r"$f(t) = t^4 - \frac{1}{3}t^3$",
            lambda t: t**4 - (t**3 / 3),
            lambda t: 4 * t**3 - t**2,
        ),
        (
            r"$f(t) = \sin(t)$",
            np.sin,
            np.cos,
        ),
        (
            r"$f(t) = e^t$",
            np.exp,
            np.exp,
        ),
        (
            r"$f(t) = \frac{1}{1 + t}$",
            lambda t: 1 / (t + 1),
            lambda t: -1 / (t + 1) ** 2,
        ),
        (
            r"$f(t) = t - t^3 + \cos(t) - e^{t/2}$",
            lambda t: t - t**3 + np.cos(t) - np.exp(t / 2),
            lambda t: 1 - 3 * t**2 - np.sin(t) - np.exp(t / 2) / 2,
        ),
    )

    # Constructor -------------------------------------------------------------
    def __init__(self, time_domain):
        """Set the time domain."""
        if not isinstance(time_domain, np.ndarray) or time_domain.ndim != 1:
            raise ValueError("time_domain must be a one-dimensional array")
        self.__t = time_domain

    # Properties --------------------------------------------------------------
    @property
    def time_domain(self):
        """Time domain of the snapshot data, a (k,) ndarray."""
        return self.__t

    def __str__(self):
        """String representation: class name, time domain."""
        out = [self.__class__.__name__]
        t = self.time_domain
        out.append(f"time_domain: {t.size} entries in [{t.min()}, {t.max()}]")
        return "\n  ".join(out)

    def __repr__(self):
        """Unique ID + string representation."""
        return utils.str2repr(self)

    # Main routine ------------------------------------------------------------
    def _check_dimensions(self, states, inputs, check_against_time=True):
        """Check dimensions and alignment of the state and inputs."""
        if states.ndim != 2:
            raise errors.DimensionalityError("states must be two-dimensional")
        if check_against_time and states.shape[-1] != self.time_domain.size:
            raise errors.DimensionalityError(
                "states not aligned with time_domain"
            )
        if inputs is not None:
            if inputs.ndim == 1:
                inputs = inputs.reshape((1, -1))
            if inputs.shape[1] != states.shape[1]:
                raise errors.DimensionalityError(
                    "states and inputs not aligned"
                )
        return states, inputs

    @abc.abstractmethod
    def estimate(self, states, inputs=None):
        """Estimate the first time derivatives of the states.

        Parameters
        ----------
        states : (r, k) ndarray
            State snapshots, either full or (preferably) reduced.
        inputs : (m, k) ndarray or None
            Inputs corresponding to the state snapshots, if applicable.

        Returns
        -------
        _states : (r, k') ndarray
            Subset of the state snapshots.
        ddts : (r, k') ndarray
            First time derivatives corresponding to ``_states``.
        _inputs : (m, k') ndarray or None
            Inputs corresponding to ``_states``, if applicable.
            **Only returned** if ``inputs`` is provided.
        """
        raise NotImplementedError  # pragma: no cover

    # Verification ------------------------------------------------------------
    def verify_shapes(self, r: int = 5, m: int = 3):
        """Verify that :meth:`estimate()` is consistent in the sense that the
        all outputs have the same number of columns. This method does **not**
        check the accuracy of the derivative estimation.

        Parameters
        ----------
        r : int
            State dimension to use in the check.
        m : int
            Number of inputs to use in the check.
            The ``inputs`` argument used to verify :meth:`estimate()` is
            a two-dimensional array of shape ``(m, k)`` even if ``m = 1``,
            where ``k`` is the size of the time domain.
        """
        # Get random data. What matters here is the size, not the entries.
        k = self.time_domain.size
        Q = np.random.random((r, k))
        U = np.random.random((m, k))

        # Call estimate() with inputs=None.
        outputs = self.estimate(Q)
        if not isinstance(outputs, tuple) or len(outputs) != 2:
            raise errors.VerificationError(
                "len(estimate(states, inputs=None)) != 2"
            )

        _Q, _dQdt = outputs
        if _Q.shape[0] != r:
            raise errors.VerificationError(
                "estimate(states)[0].shape[0] != states.shape[0]"
            )
        if _dQdt.shape[0] != r:
            raise errors.VerificationError(
                "estimate(states)[1].shape[0] != states.shape[0]"
            )
        if _Q.shape[1] != _dQdt.shape[1]:
            print(_Q.shape[1], _dQdt.shape[1])
            raise errors.VerificationError(
                "Q.shape[1] != dQdt.shape[1] "
                "where Q, dQdt = estimate(states, inputs=None)"
            )

        # Call estimate() with non-None inputs.
        outputs = self.estimate(Q, U)
        if not isinstance(outputs, tuple) or len(outputs) != 3:
            raise errors.VerificationError(
                "len(estimate(states, inputs)) != 3"
            )

        _Q, _dQdt, _U = outputs
        if _U is None:
            raise errors.VerificationError(
                "estimates(states, inputs)[2] should not be None"
            )
        if _U.shape[0] != m:
            raise errors.VerificationError(
                "estimate(states, inputs)[2].shape[0] != inputs.shape[0]"
            )
        if _U.shape[1] != _Q.shape[1]:
            raise errors.VerificationError(
                "Q.shape[1] != U.shape[1] "
                "where Q, _, U = estimate(states, inputs)"
            )

        print("estimate() output shapes are consistent")

    def verify(self, plot: bool = False, return_errors=False):
        """Verify that :meth:`estimate()` is consistent in the sense that the
        all outputs have the same number of columns and test the accuracy of
        the results on a few test problems.

        Parameters
        ----------
        plot : bool
            If ``True``, plot the relative errors of the derivative estimation
            errors as a function of the time step.
            If ``False`` (default), print a report of the relative errors.
        return_errors : bool
            If ``True``, return the errors for each test as a dictionary.
            If ``False`` (default), return nothing.

        Returns
        -------
        errors : dict
            Estimation errors for each test case.
            Time steps are listed as ``errors[dts]``.
            **Only returned** if ``return_errors=True``.
        """
        self.verify_shapes()

        time_domain = self.time_domain  # Record original time domain.
        dts = np.logspace(-12, -1, 12)[::-1]
        estimation_errors = collections.defaultdict(list)
        estimation_errors["dts"] = dts

        t_base = np.arange(1001)
        for dt in dts:
            # Construct test cases.
            t = 1 + (dt * t_base)
            Q = np.array([test[1](t) for test in self.__tests])
            dQdt = np.array([test[2](t) for test in self.__tests])
            self.__t = t

            # Call the derivative estimator.
            Q_est, dQdt_est = self.estimate(Q, None)

            # Use new states to infer the indices of the returns.
            start = np.argmin(
                la.norm(Q - Q_est[:, 0].reshape((-1, 1))), axis=0
            )
            s = slice(start, start + Q_est.shape[1])
            truth = dQdt[:, s]

            # Calculate the relative error of the time derivative estimate.
            denom = la.norm(truth, axis=1)
            errs = la.norm(dQdt_est - truth, axis=1) / denom
            for test, err in zip(self.__tests, errs):
                estimation_errors[test[0]].append(err)

        if plot:
            _, ax = plt.subplots(1, 1)
            for test in self.__tests:
                name = test[0]
                ax.loglog(
                    dts,
                    estimation_errors[name],
                    ".-",
                    linewidth=0.5,
                    markersize=5,
                    label=name,
                )
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(bottom=max(ymin, 1e-14), top=min(ymax, 10))
            ax.set_xlabel("time step")
            ax.set_ylabel("relative error")
            ax.legend(ncols=2, frameon=False, fontsize="small")
            ax.set_title("Time derivative estimation errors")
        else:
            print(
                (title := "\nTime derivative estimation relative errors"),
                "=" * len(title),
                sep="\n",
            )
            for test in self.__tests:
                name = test[0]
                rawname = name.strip("$")
                print(f"\n{rawname}", len(rawname) * "-", sep="\n")
                for dt, err in zip(dts, estimation_errors[name]):
                    print(f"dt = {dt:.1e}:\terror = {err:.4e}")

        self.__t = time_domain  # Restore original time domain.
        if return_errors:
            return estimation_errors



# utils/_repr.py
"""Canonical string representation for objects with a ``__str__()`` method."""

__all__ = [
    "str2repr",
]


def str2repr(obj) -> str:
    """
    Canonical string representation for objects with a ``__str__()`` method.
    """
    uniqueID = f"<{obj.__class__.__name__} object at {hex(id(obj))}>"
    return f"{uniqueID}\n{str(obj)}"



# _gridsearch.py
"""Routine for performing a grid search, followed by a derivative-free
optimization.
"""

__all__ = [
    "gridsearch",
    "DiscreteRegTest",
    "ContinuousRegTest",
]

import abc
import typing
import warnings
import dataclasses
import numpy as np
import scipy.optimize

from .. import errors


__MAXOPTVAL = 1e20


def gridsearch(
    func: typing.Callable,
    candidates: np.ndarray,
    gridsearch_only: bool = False,
    label: str = "parameter",
    verbose: bool = False,
) -> typing.Union[float, np.ndarray]:
    r"""Minimize a function by first checking a collection of candidates, then
    following up with a derivative-free optimization routine.

    If the candidates are one-dimensional, meaning ``func()`` is a scalar map
    :math:`f:\RR_+\to\RR`, the optimization is carried out via
    :func:`scipy.optimize.minimize_scalar` with ``method='brent'``.

    Otherwise (:math:`f:\RR_+^p\to\RR` for some :math:`p > 1`), the
    optimization is carried out via :func:`scipy.optimize.minimize` with
    ``method='Nelder-Mead'``.

    Parameters
    ----------
    func : callable(candidate) -> float
        Function to minimize.
    candidates : list of floats or ndarrays
        Candidates to check before starting the optimization routine.
    gridsearch_only : bool
        If ``True``, stop after checking all ``candidates`` and do not follow
        up with optimization.
    label : str
        Label for the types of candidates being checked.
    verbose : bool
        If ``True``, print information at each iteration.

    Returns
    -------
    winner : float or ndarray
        Optimized candidate.

    Raises
    ------
    RuntimeError
        If the grid search fails because ``func()`` returns ``np.inf`` for
        each candidate.

    Warns
    -----
    OpInfWarning
        1) If the ``candidates`` are one-dimensional and the grid search winner
        is the smallest or largest candidate, or 2) If the minimization fails
        or does not result in a solution that improves upon the grid search.

    Notes
    -----
    The optimization minimizes ``func(x)`` by varying ``x`` logarithmically.
    Unless ``gridsearch_only=True``, it is assumed that all entries of the
    argument ``x`` are positive.

    If only one scalar candidate is given, the (one-dimensional) optimization
    sets the bounds to ``[candidate / 100, candidate * 100]``.
    """
    # Process the candidates.
    candidates = np.atleast_1d(candidates)
    if np.shape(candidates[0]) == (1,):
        candidates = [reg[0] for reg in candidates]
    scalar_regularization = np.ndim(candidates[0]) == 0

    def pstr(params):
        if np.isscalar(params):
            return f"{params:.4e}"
        else:
            return "[" + ", ".join([f"{p:.3e}" for p in params]) + "]"

    def linfunc(params):
        if verbose:
            print(f"{label.title()} {pstr(params)}...", end="", flush=True)
        out = func(params)
        if verbose:
            if out == np.inf:
                print("UNSTABLE")
            else:
                print(f"{out:.8%} error")
        return out

    def logfunc(log10params):
        params = 10**log10params
        return min(__MAXOPTVAL, linfunc(params))

    # Grid search.
    num_tests = len(candidates)
    if verbose:
        print(f"\nGRID SEARCH ({num_tests} candidates)")
        ndigits = len(str(num_tests))
    winning_error, winner_index = np.inf, None
    if scalar_regularization:
        candidates = np.sort(candidates)
    for i, candidate in enumerate(candidates):
        if verbose:
            print(f"({i+1: >{ndigits}d}/{num_tests:d}) ", end="")
        if (error := linfunc(candidate)) < winning_error:
            winning_error = error
            winner_index = i
    if winner_index is None:
        raise RuntimeError(f"{label} grid search failed")
    gridsearch_winner = candidates[winner_index]
    if verbose:
        print(
            f"Best {label} candidate via grid search:",
            pstr(gridsearch_winner),
        )
    if gridsearch_only:
        return gridsearch_winner

    # Post-process results if the candidates are scalars.
    if scalar_regularization:
        if num_tests == 1:
            search_bounds = [gridsearch_winner / 1e2, 1e2 * gridsearch_winner]
        elif winner_index == 0:
            warnings.warn(
                f"smallest {label} candidate won grid search, "
                f"consider using smaller candidates",
                errors.OpInfWarning,
            )
            search_bounds = [gridsearch_winner / 1e2, candidates[1]]
        elif winner_index == num_tests - 1:
            warnings.warn(
                f"largest {label} candidate won grid search, "
                f"consider using larger candidates",
                errors.OpInfWarning,
            )
            search_bounds = [candidates[-2], 1e2 * gridsearch_winner]
        else:
            search_bounds = [
                candidates[winner_index - 1],
                candidates[winner_index + 1],
            ]

    # Follow up grid search with minimization-based search.
    if verbose:
        print("OPTIMIZATION")
    if scalar_regularization:
        opt_result = scipy.optimize.minimize_scalar(
            logfunc,
            method="bounded",
            bounds=np.log10(search_bounds),
        )
    else:
        opt_result = scipy.optimize.minimize(
            logfunc,
            x0=np.log10(gridsearch_winner),
            method="Nelder-Mead",
        )

    # Report results.
    success = opt_result.success and opt_result.fun != __MAXOPTVAL
    if not success or (winning_error < opt_result.fun):
        warnings.warn(
            f"{label} grid search performed better than optimization, "
            "falling back on grid search solution",
            errors.OpInfWarning,
        )
        return gridsearch_winner

    optimization_winner = 10**opt_result.x

    if verbose:
        print(
            f"Best {label} candidate via optimization:",
            pstr(optimization_winner),
        )
    return optimization_winner


# Additional regularization test cases ========================================
class _RegTest(abc.ABC):
    """Base class for regularization selection test cases."""

    def evaluate(self, model, **predict_options):
        """Evaluate the ``model`` under the test case conditions.

        Parameters
        ----------
        model : opinf.models object
            Instantiated model, ready for prediction.
        predict_options : dict
            Additional keyword arguments for ``model.predict()``.

        Returns
        -------
        result : bool
            ``True`` if the ``model`` produces a stable solution under the test
            case conditions, ``False`` otherwise.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            solution = model.predict(*self.predict_args, **predict_options)
        return not self.unstable(solution)

    @abc.abstractmethod
    def unstable(self, Q):
        """Return ``True`` if the trajectory ``Q`` is unstable."""
        pass  # pragma: no cover

    @abc.abstractmethod
    def copy(self, newICs):
        """Return a copy of this test case with new initial conditions."""
        pass  # pragma: no cover


@dataclasses.dataclass(frozen=True)
class DiscreteRegTest(_RegTest):
    """Test case for regularization selection with fully discrete models.

    The ``test_cases`` argument of
    :meth:`opinf.roms.ROM.fit_regselect_discrete` is a list of these.

    Parameters
    ----------
    initial_conditions : (n,) ndarray
        Initial conditions to be tested.
    niters : int
        Number of iterations to step forward from the initial conditions.
    parameters : (p,) ndarray or float or None
        Parameter value to be tested.
    inputs : (m, niters) ndarray or None
        Inputs to use in the forward prediction, if the model takes inputs.
    bound : float or None
        Amount that the forward prediction is allowed to deviate from the
        initial conditions without the trajectory being classified as unstable.
    """

    initial_conditions: np.ndarray
    niters: int
    parameters: np.ndarray = None
    inputs: np.ndarray = None
    bound: float = None
    predict_args: tuple = dataclasses.field(init=False)

    def __post_init__(self):
        predict_args = [
            self.initial_conditions,
            self.niters,
            self.inputs,
        ]
        if self.parameters is not None:
            predict_args.insert(0, self.parameters)
        object.__setattr__(self, "predict_args", tuple(predict_args))

    def copy(self, newICs):
        return self.__class__(
            newICs,
            self.niters,
            self.parameters,
            self.inputs,
            self.bound,
        )

    def unstable(self, Q):
        """Return ``True`` if the trajectory ``Q`` is unstable."""
        if np.isnan(Q).any() or np.isinf(Q).any():
            return True
        if (B := self.bound) is not None:
            Qshifted = Q - self.initial_conditions.reshape((-1, 1))
            if np.any(np.abs(Qshifted) > B):
                return True
        return False


@dataclasses.dataclass(frozen=True)
class ContinuousRegTest(_RegTest):
    """Test case for regularization selection with time-continuous models.

    The ``test_cases`` argument of
    :meth:`opinf.roms.ROM.fit_regselect_continuous` is a list of these.

    Parameters
    ----------
    initial_conditions : (n,) ndarray
        Initial conditions to be tested.
    time_domain : (k,) ndarray
        Time domain over which to solve the model forward in time.
    parameters : (p,) ndarray or float or None
        Parameter value to be tested.
    inputs : callable or None
        Input function, mapping time to input vectors, to use in the forward
        prediction, if the model takes inputs.
    bound : float or None
        Amount that the forward prediction is allowed to deviate from the
        initial conditions without the trajectory being classified as unstable.
    """

    initial_conditions: np.ndarray
    time_domain: np.ndarray
    parameters: np.ndarray = None
    input_function: typing.Callable = None
    bound: float = None
    predict_args: tuple = dataclasses.field(init=False)

    def __post_init__(self):
        predict_args = [
            self.initial_conditions,
            self.time_domain,
            self.input_function,
        ]
        if self.parameters is not None:
            predict_args.insert(0, self.parameters)
        object.__setattr__(self, "predict_args", tuple(predict_args))

    def copy(self, newICs):
        return self.__class__(
            newICs,
            self.time_domain,
            self.parameters,
            self.input_function,
            self.bound,
        )

    def unstable(self, Q):
        """Return ``True`` if the trajectory ``Q`` is unstable."""
        if Q.shape[-1] != self.time_domain.size:
            return True
        if np.isnan(Q).any() or np.isinf(Q).any():
            return True
        if (B := self.bound) is not None:
            Qshifted = Q - self.initial_conditions.reshape((-1, 1))
            if np.any(np.abs(Qshifted) > B):
                return True
        return False



# lift/_polynomial.py
r"""Polynomial lifting maps, e.g., :math:`q \to (q, q^2, ...)`."""

__all__ = [
    "QuadraticLifter",
    "PolynomialLifter",
]

import numbers
import warnings
import numpy as np

from .. import errors
from ._base import LifterTemplate


class QuadraticLifter(LifterTemplate):
    r"""Quadratic lifting map :math:`q \to (q, q^2)`."""

    @staticmethod
    def lift(states):
        r"""Apply the lifting map :math:`q \to (q, q^2)`.

        Parameters
        ----------
        states : (n, k) ndarray
            Native state variables.

        Returns
        -------
        lifted_states : (2n, k) ndarray
            Learning variables.
        """
        return np.concatenate((states, states**2))

    @staticmethod
    def lift_ddts(states, ddts):
        r"""Get the time derivatives of the lifted variables,
        :math:`\frac{\partial}{\partial t}(q, q^2) = (q_t, 2qq_t)`.

        Parameters
        ----------
        states : (n, k) ndarray
            Native state variables.
        ddts : (n, k) ndarray
            Time derivatives of the native state variables. Each column
            ``ddts[:, j]`` corresponds to the state vector ``states[:, j]``.

        Returns
        -------
        ddts : (2n, k) ndarray
            Time derivatives of the learning variables.
        """
        return np.concatenate((ddts, 2 * states * ddts))

    @staticmethod
    def unlift(lifted_states):
        r"""Apply the reverse lifting map :math:`(q, q^2) \to q`.

        Parameters
        ----------
        lifted_states : (2n, k) ndarray
            Learning variables.

        Returns
        -------
        states : (n, k) ndarray
            Native state variables.
        """
        return np.split(lifted_states, 2, axis=0)[0]


class PolynomialLifter(LifterTemplate):
    r"""Polynomial lifting map :math:`q \to (q, q^2, q^3, ...)`.

    Parameters
    ----------
    orders : tuple
        Polynomial orders in the learning variables. For example,
        ``orders=(1, 2, 4)`` means the lifting transformation is given by
        :math:`q \to (q, q^2, q^4)`. The orders need not be positive integers,
        e.g., ``orders=(-1, 0.5)`` indicates :math:`q \to (1/q, \sqrt{q})`.
    """

    def __init__(self, orders: tuple):
        """Set the polynomial orders."""
        self.orders = orders

    # Properties --------------------------------------------------------------
    @property
    def orders(self) -> tuple:
        r"""Polynomial orders in the learning variables. For example,
        ``orders=(1, 2, 4)`` means the lifting transformation is given by
        :math:`q \to (q, q^2, q^4)`. The orders need not be positive integers,
        e.g., ``orders=(-1, 0.5)`` indicates :math:`q \to (1/q, \sqrt{q})`.
        """
        return self.__orders

    @orders.setter
    def orders(self, ps: tuple):
        """Set the polynomial orders."""
        if isinstance(ps, numbers.Number):
            ps = (int(ps),)
        for p in ps:
            if not isinstance(p, numbers.Number):
                raise TypeError("'orders' must be a sequence of numbers")
        if len(ps) == 1 and ps[0] == 0:
            warnings.warn(
                "q -> q^0 = 1 is not invertible",
                errors.OpInfWarning,
            )

        self.__orders = tuple(ps)
        self.__nvars = len(self.__orders)

    @property
    def num_variables(self) -> int:
        """Number of learning variables."""
        return self.__nvars

    def __str__(self) -> str:
        """String representation: lifting map description"""
        variables = ", ".join(
            [(f"q^{p}" if p != 1 else "q") for p in self.__orders]
        )
        if self.num_variables > 1:
            variables = "(" + variables + ")"
        return f"Lifting map q -> {variables}"

    # Lifting map -------------------------------------------------------------
    def lift(self, states):
        r"""Apply the lifting map :math:`q \to (q, q^2, ...)`.

        Parameters
        ----------
        states : (n, k) ndarray
            Native state variables.

        Returns
        -------
        lifted_states : (num_variables * n, k) ndarray
            Learning variables.
        """
        return np.concatenate([states**p for p in self.__orders])

    def lift_ddts(self, states, ddts):
        """Get the time derivatives of the lifted variables,
        :math:`(q_t, 2qq_t, ...)`.

        Parameters
        ----------
        states : (n, k) ndarray
            Native state variables.
        ddts : (n, k) ndarray
            Time derivatives of the native state variables. Each column
            ``ddts[:, j]`` corresponds to the state vector ``states[:, j]``.

        Returns
        -------
        ddts : (num_variables * n, k) ndarray
            Time derivatives of the learning variables.
        """
        return np.concatenate(
            [p * states ** (p - 1) * ddts for p in self.__orders]
        )

    def unlift(self, lifted_states):
        r"""Apply the reverse lifting map :math:`(q, q^2, ...) \to q`.

        Parameters
        ----------
        lifted_states : (num_variables * n, k) ndarray
            Learning variables.

        Returns
        -------
        states : (n, k) ndarray
            Native state variables.
        """
        variables = np.split(lifted_states, self.num_variables, axis=0)
        if 1 in self.__orders:
            return variables[self.__orders.index(1)]
        for var, p in zip(variables, self.__orders):
            if p != 0:
                return var ** (1 / p)
        raise ZeroDivisionError("q -> q^0 = 1 is not invertible")



# operators/_utils.py
"""Private utility functions for working with Operator classes."""

__all__ = [
    "has_inputs",
    "is_nonparametric",
    "is_parametric",
    "is_opinf",
    "is_uncalibrated",
    "is_affine",
    "is_interpolated",
    "nonparametric_to_affine",
    "nonparametric_to_interpolated",
]

from ._base import (
    has_inputs,
    is_nonparametric,
    is_parametric,
    is_opinf,
    is_uncalibrated,
)
from ._affine import is_affine, nonparametric_to_affine
from ._interpolate import is_interpolated, nonparametric_to_interpolated



# operators/_interpolate.py
"""Classes for OpInf parametric operators where the parametric dependence is
handled with element-wise interpolation.
"""

__all__ = [
    "InterpConstantOperator",
    "InterpLinearOperator",
    "InterpQuadraticOperator",
    "InterpCubicOperator",
    "InterpInputOperator",
    "InterpStateInputOperator",
    # Deprecations:
    "InterpolatedConstantOperator",
    "InterpolatedLinearOperator",
    "InterpolatedQuadraticOperator",
    "InterpolatedCubicOperator",
    "InterpolatedInputOperator",
    "InterpolatedStateInputOperator",
]

import warnings
import numpy as np
import scipy.linalg as la
import scipy.interpolate as spinterp

from .. import errors, utils
from ._base import ParametricOpInfOperator, InputMixin
from ._nonparametric import (
    ConstantOperator,
    LinearOperator,
    QuadraticOperator,
    CubicOperator,
    InputOperator,
    StateInputOperator,
)


# Base class ==================================================================
class _InterpOperator(ParametricOpInfOperator):
    r"""Base class for parametric operators where the parameter dependence
    is handled with element-wise interpolation.

    For a set of training parameter values :math:`\{\bfmu_i\}_{i=0}^{s-1}`,
    this type of operator is given by
    :math:`\Ophat_\ell(\qhat, \u, \bfmu) = \Ohat_\ell(\bfmu)\d(\qhat, \u)`
    where :math:`\Ohat_{\ell}(\bfmu)` is calculated by interpolating the
    operator matrix entries that correspond to each parameter value:

    .. math::
       \Ohat_{\ell}(\bfmu)
        = \textrm{interpolate}(
       (\bfmu_0,\Ohat_{\ell}^{(0)}),
       \ldots,(\Ohat_{\ell}^{(s-1)}\bfmu_{s-1});\bfmu),

    where :math:`\Ohat_\ell^{(i)} = \Ohat_\ell(\bfmu_i)` for each
    :math:`i=0,\ldots,s-1`.

    Parent class: :class:`opinf.operators.ParametricOpInfOperator`

    Parameters
    ----------
    training_parameters : list of s scalars or (p,) 1D ndarrays
        Parameter values for which the operator matrix is known
        or will be inferred from data. If not provided in the constructor,
        use :meth:`set_training_parameters` later.
    entries : list of s ndarrays, or None
        Operator matrices corresponding to the ``training_parameters``.
        If not provided in the constructor, use :meth:`set_entries` later.
    InterpolatorClass : type or None
        Class for the elementwise interpolation. Must obey the syntax

            >>> interpolator = InterpolatorClass(data_points, data_values)
            >>> interpolator_evaluation = interpolator(new_data_point)

        This can be, e.g., a class from :mod:`scipy.interpolate`.
        If ``None`` (default), use :class:`scipy.interpolate.CubicSpline`
        for one-dimensional parameters and
        :class:`scipy.interpolate.LinearNDInterpolator` otherwise.
        If not provided in the constructor, use :meth:`set_interpolator` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    # Initialization ----------------------------------------------------------
    def __init__(
        self,
        training_parameters=None,
        entries=None,
        InterpolatorClass: type = None,
        fromblock=False,
    ):
        """Set attributes and, if training parameters and entries are given,
        construct the elementwise operator matrix interpolator.
        """
        ParametricOpInfOperator.__init__(self)

        self.__parameters = None
        self.__interpolator = None
        self.__InterpolatorClass = InterpolatorClass

        if training_parameters is not None:
            self.set_training_parameters(training_parameters)
        if entries is not None:
            self.set_entries(entries, fromblock=fromblock)

    @classmethod
    def _from_operators(
        cls,
        training_parameters,
        operators,
        InterpolatorClass: type = None,
    ):
        """Interpolate an existing set of nonparametric operators with
        populated entries.

        Parameters
        ----------
        operators : list of nonparametric :mod:`opinf.operators` operators
            Operators to interpolate with ``entries`` already set.
        """
        # Check everything is initialized.
        for op in operators:
            if not isinstance(op, cls._OperatorClass):
                raise TypeError(
                    "can only interpolate operators of type "
                    f"'{cls._OperatorClass.__name__}'"
                )
            if op.entries is None:
                raise ValueError(
                    "operators must have entries set in order to interpolate"
                )

        # Extract the entries.
        return cls(
            training_parameters,
            InterpolatorClass=InterpolatorClass,
            entries=[op.entries for op in operators],
            fromblock=False,
        )

    def _clear(self) -> None:
        """Reset the operator to its post-constructor state without entries."""
        ParametricOpInfOperator._clear(self)
        self.__interpolator = None

    # Properties --------------------------------------------------------------
    @property
    def training_parameters(self):
        """Parameter values where the operator matrix is known
        or will be inferred from data.
        """
        return self.__parameters

    @training_parameters.setter
    def training_parameters(self, params):
        """Set the training parameter values."""
        self.set_training_parameters(params)

    def set_training_parameters(self, training_parameters):
        """Set the training parameter values.

        Parameters
        ----------
        training_parameters : list of s scalars or (p,) 1D ndarrays
            Parameter values for which the operator matrix is known
            or will be inferred from data.
        """
        if self.__interpolator is not None:
            raise AttributeError(
                "can't set attribute (entries already set, call _clear())"
            )

        # Check argument dimensions.
        self._check_shape_consistency(
            training_parameters,
            "training_parameters",
        )

        parameters = np.array(training_parameters)
        if parameters.ndim not in (1, 2):
            raise ValueError("parameter values must be scalars or 1D arrays")
        self._set_parameter_dimension_from_values(parameters)
        if parameters.ndim == 2 and parameters.shape[-1] == 1:
            parameters = parameters.ravel()
        self.__parameters = parameters

    @property
    def entries(self) -> np.ndarray:
        """Operator matrices corresponding to the training parameters values,
        i.e., ``entries[i]`` is the operator matrix corresponding to the
        parameter value ``training_parameters[i]``.
        """
        return ParametricOpInfOperator.entries.fget(self)

    def set_entries(self, entries, fromblock: bool = False) -> None:
        r"""Set the operator matrices at the training parameter values.

        Parameters
        ----------
        entries : list of s (r, d) ndarrays, or (r, sd) ndarray
            Operator matrices, either as a list of arrays
            (``fromblock=False``, default)
            or as a horizontal concatenatation of arrays (``fromblock=True``).
        fromblock : bool
            If ``True``, interpret ``entries`` as a horizontal concatenation
            of arrays; if ``False`` (default), interpret ``entries`` as a list
            of arrays.
        """
        if self.training_parameters is None:
            raise AttributeError(
                "training_parameters have not been set, "
                "call set_training_parameters() first"
            )

        # Extract / verify the entries.
        n_params = len(self.__parameters)
        if fromblock:
            if entries.ndim not in (1, 2):
                raise ValueError(
                    "entries must be a 1- or 2-dimensional ndarray "
                    "when fromblock=True"
                )
            entries = np.split(entries, n_params, axis=-1)
        if np.ndim(entries) > 1:
            self._check_shape_consistency(entries, "entries")
        if (n_arrays := len(entries)) != n_params:
            raise ValueError(
                f"{n_params} = len(training_parameters) "
                f"!= len(entries) = {n_arrays}"
            )

        ParametricOpInfOperator.set_entries(
            self,
            np.array([self._OperatorClass(A).entries for A in entries]),
        )
        self.set_interpolator(self.__InterpolatorClass)

    # Interpolation -----------------------------------------------------------
    @property
    def interpolator(self):
        """Interpolator object for evaluating the operator at specified
        parameter values.
        """
        return self.__interpolator

    def set_interpolator(self, InterpolatorClass):
        """Construct the interpolator for the operator matrix.

        Parameters
        ----------
        InterpolatorClass : type
            Class for the elementwise interpolation. Must obey the syntax

               >>> interpolator = InterpolatorClass(data_points, data_values)
               >>> interpolator_evaluation = interpolator(new_data_point)

            This can be, e.g., a class from :mod:`scipy.interpolate`.
        """
        if self.entries is not None:
            params = self.training_parameters
            entries = self.entries

            # Default interpolator classes.
            if InterpolatorClass is None:
                if (dim := params.ndim) == 1:
                    InterpolatorClass = spinterp.CubicSpline
                    paramsort = np.argsort(params)
                    params = params[paramsort]
                    entries = self.entries[paramsort]
                elif dim == 2:
                    InterpolatorClass = spinterp.LinearNDInterpolator

            # Do the interpolation.
            self.__interpolator = InterpolatorClass(params, entries)

        self.__InterpolatorClass = InterpolatorClass

    # Magic methods -----------------------------------------------------------
    def __eq__(self, other) -> bool:
        """Test whether the training parameters and operator matrices of two
        _InterpOperator objects are the same.
        """
        if not isinstance(other, self.__class__):
            return False
        if (
            self.training_parameters is None
            and other.training_parameters is not None
        ) or (
            self.training_parameters is not None
            and other.training_parameters is None
        ):
            return False
        if self.training_parameters is not None:
            if (
                self.training_parameters.shape
                != other.training_parameters.shape
            ):
                return False
            if not np.all(
                self.training_parameters == other.training_parameters
            ):
                return False
        if self.__InterpolatorClass is not other.__InterpolatorClass:
            return False
        if (self.entries is None and other.entries is not None) or (
            self.entries is not None and other.entries is None
        ):
            return False
        if self.entries is not None:
            if self.shape != other.shape:
                return False
            return np.allclose(self.entries, other.entries)
        return True

    def __str__(self):
        lines = ParametricOpInfOperator.__str__(self).split("\n")

        nparams = "None"
        if (params := self.training_parameters) is not None:
            nparams = len(params)
        lines.insert(-1, f"  training parameters: {nparams}")

        ICname = "None"
        if (IC := self.__InterpolatorClass) is not None:
            ICname = IC.__name__
        lines.insert(-1, f"  type(interpolator):  {ICname}")

        return "\n".join(lines)

    # Evaluation --------------------------------------------------------------
    @utils.requires("entries")
    def evaluate(self, parameter):
        r"""Evaluate the operator at the given parameter value.

        Parameters
        ----------
        parameter : (p,) ndarray or float
            Parameter value :math:`\bfmu` at which to evalute the operator.

        Returns
        -------
        op : nonparametric :mod:`opinf.operators` operator
            Nonparametric operator corresponding to the parameter value.
        """
        self._check_parametervalue_dimension(parameter)
        if self.parameter_dimension == 1 and not np.isscalar(parameter):
            parameter = parameter[0]
        entries = self.interpolator(parameter)
        if entries.ndim == 3:
            entries = entries[0]
        return self._OperatorClass(entries)

    # Dimensionality reduction ------------------------------------------------
    @utils.requires("entries")
    def galerkin(self, Vr, Wr=None):
        r"""Project this operator to a low-dimensional linear space.

        Consider an interpolatory operator

        .. math::
           \f_\ell(\q,\u;\bfmu)
           = \textrm{interpolate}(
           (\bfmu_0,\f_{\ell}^{(0)}(\q,\u)),\ldots,
           (\bfmu_{s-1},\f_{\ell}^{(s-1)}(\q,\u)); \bfmu),

        where

        * :math:`\q\in\RR^n` is the full-order state,
        * :math:`\u\in\RR^m` is the input,
        * :math:`\bfmu_0,\ldots,\bfmu_{s-1}\in\RR^p`
          are the (fixed) training parameter values,
        * :math:`\f_{\ell}^{(i)}(\q,\u) = \f_{\ell}(\q,\u;\bfmu_i)`
          is the operators evaluated at the :math:`i`-th training parameter
          values, :math:`i=1,\ldots,s`, and
        * :math:`\bfmu\in\RR^p` is a new parameter value
          at which to evaluate the operator.

        Given a *trial basis* :math:`\Vr\in\RR^{n\times r}` and a *test basis*
        :math:`\Wr\in\RR^{n\times r}`, the corresponding *intrusive projection*
        of :math:`\f` is the interpolatory operator

        .. math::
           \fhat_{\ell}(\qhat,\u;\bfmu)
           = \textrm{interpolate}(
           (\bfmu_0,\Wr\trp\f_{\ell}^{(0)}(\Vr\qhat,\u)),\ldots,
           (\bfmu_{s-1},\Wr\trp\f_{\ell}^{(s-1)}(\Vr\qhat,\u)); \bfmu),

        Here, :math:`\qhat\in\RR^r` is the reduced-order state, which enables
        the low-dimensional state approximation :math:`\q = \Vr\qhat`.
        If :math:`\Wr = \Vr`, the result is called a *Galerkin projection*.
        If :math:`\Wr \neq \Vr`, it is called a *Petrov-Galerkin projection*.

        Parameters
        ----------
        Vr : (n, r) ndarray
            Basis for the trial space.
        Wr : (n, r) ndarray or None
            Basis for the test space. If ``None``, defaults to ``Vr``.

        Returns
        -------
        op : operator
            New object of the same class as ``self``.
        """
        return self.__class__(
            training_parameters=self.training_parameters,
            entries=[
                self._OperatorClass(A).galerkin(Vr, Wr).entries
                for A in self.entries
            ],
            InterpolatorClass=self.__InterpolatorClass,
            fromblock=False,
        )

    # Operator inference ------------------------------------------------------
    @classmethod
    def datablock(cls, parameters, states, inputs=None) -> np.ndarray:
        r"""Return the data matrix block corresponding to the operator.

        For interpolated operators, this is a block diagonal matrix where the
        :math:`i`-th block is the data block for the corresponding
        nonparametric operator class with the state snapshots and inputs
        for training parameter :math:`\bfmu_i`.

        Parameters
        ----------
        parameters : (s, p) ndarray
            Traning parameter values :math:`\bfmu_{0},\ldots,\bfmu_{s-1}`.
        states : list of s (r, k) or (k,) ndarrays
            State snapshots for each of the `s` training parameter values.
            If each snapshot matrix is 1D, it is assumed that :math:`r = 1`.
        inputs : list of s (m, k) or (k,) ndarrays
            Inputs corresponding to the state snapshots.
            If each input matrix is 1D, it is assumed that :math:`m = 1`.

        Returns
        -------
        block : (sd, sk) ndarray
            Data block for the interpolated operator. Here, `d` is the number
            of rows in the data block corresponding to a single training
            parameter value.
        """
        if not issubclass(cls, InputMixin):
            inputs = [None] * len(parameters)
        return la.block_diag(
            *[
                cls._OperatorClass.datablock(Q, U)
                for Q, U in zip(states, inputs)
            ]
        )

    @classmethod
    def operator_dimension(cls, s: int, r: int, m: int) -> int:
        r"""Number of columns in the concatenated operator matrix.

        Parameters
        ----------
        s : int
            Number of training parameter values.
        r : int
            State dimension.
        m : int or None
            Input dimension.
        """
        return s * cls._OperatorClass.operator_dimension(r, m)

    # Model persistence -------------------------------------------------------
    def copy(self):
        """Return a copy of the operator."""
        params = self.training_parameters
        return self.__class__(
            training_parameters=params.copy() if params is not None else None,
            InterpolatorClass=self.__InterpolatorClass,
            entries=self.entries.copy() if self.entries is not None else None,
        )

    def save(self, savefile: str, overwrite: bool = False) -> None:
        """Save the operator to an HDF5 file.

        If the :attr:`interpolator` is not from :mod:`scipy.interpolate`,
        it must be passed to :meth:`load()` when recovering the operator.

        Parameters
        ----------
        savefile : str
            Path of the file to save the basis in.
        overwrite : bool
            If ``True``, overwrite the file if it already exists. If ``False``
            (default), raise a ``FileExistsError`` if the file already exists.
        """
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            meta = hf.create_dataset("meta", shape=(0,))
            meta.attrs["class"] = self.__class__.__name__

            InterpolatorClassName = (
                "NoneType"
                if self.__InterpolatorClass is None
                else self.__InterpolatorClass.__name__
            )
            meta.attrs["InterpolatorClass"] = InterpolatorClassName
            if InterpolatorClassName != "NoneType" and not hasattr(
                spinterp, InterpolatorClassName
            ):
                warnings.warn(
                    "cannot serialize InterpolatorClass "
                    f"'{InterpolatorClassName}', must pass in the class "
                    "when calling load()",
                    errors.OpInfWarning,
                )

            if self.training_parameters is not None:
                hf.create_dataset(
                    "training_parameters", data=self.training_parameters
                )
            if self.entries is not None:
                hf.create_dataset("entries", data=self.entries)

    @classmethod
    def load(cls, loadfile: str, InterpolatorClass: type = None):
        """Load a parametric operator from an HDF5 file.

        Parameters
        ----------
        loadfile : str
            Path to the file where the operator was stored via :meth:`save()`.
        InterpolatorClass : type or None
            Class for the elementwise interpolation. Must obey the syntax

               >>> interpolator = InterpolatorClass(data_points, data_values)
               >>> interpolator_evaluation = interpolator(new_data_point)

            Not required if the saved operator utilizes a class from
            :mod:`scipy.interpolate`.

        Returns
        -------
        op : _InterpOperator
            Initialized operator object.
        """
        with utils.hdf5_loadhandle(loadfile) as hf:
            ClassName = hf["meta"].attrs["class"]
            if ClassName != cls.__name__:
                raise TypeError(
                    f"file '{loadfile}' contains '{ClassName}' "
                    f"object, use '{ClassName}.load()'"
                )

            # Get the InterpolatorClass.
            SavedClassName = hf["meta"].attrs["InterpolatorClass"]
            if InterpolatorClass is None:
                # Load from scipy.interpolate.
                if hasattr(spinterp, SavedClassName):
                    InterpolatorClass = getattr(spinterp, SavedClassName)
                elif SavedClassName != "NoneType":
                    raise ValueError(
                        f"unknown InterpolatorClass '{SavedClassName}', "
                        f"call {ClassName}.load({loadfile}, {SavedClassName})"
                    )
            else:
                # Warn the user if the InterpolatorClass does not match.
                if SavedClassName != (
                    InterpolatorClassName := InterpolatorClass.__name__
                ):
                    warnings.warn(
                        f"InterpolatorClass={InterpolatorClassName} does not "
                        f"match loadfile InterpolatorClass '{SavedClassName}'",
                        errors.OpInfWarning,
                    )

            return cls(
                training_parameters=(
                    hf["training_parameters"][:]
                    if "training_parameters" in hf
                    else None
                ),
                entries=(hf["entries"][:] if "entries" in hf else None),
                InterpolatorClass=InterpolatorClass,
                fromblock=False,
            )


# Public interpolated operator classes ========================================
class InterpConstantOperator(_InterpOperator):
    r"""Parametric constant operator
    :math:`\Ophat_{\ell}(\qhat,\u;\bfmu) = \chat(\bfmu) \in \RR^r`
    where the parametric dependence is handled with elementwise interpolation.

    .. math::
       \chat(\bfmu) = \textrm{interpolate}(
       (\bfmu_0,\chat^{(0)}),\ldots,(\bfmu_{s-1},\chat^{(s-1)}); \bfmu)

    Here,

    * :math:`\bfmu_0,\ldots,\bfmu_{s-1}\in\RR^p`
      are the (fixed) training parameter values, and
    * :math:`\chat^{(i)} = \chat(\bfmu_i) \in \RR^r`
      is the operator vector evaluated at the training parameter values.

    See :class:`opinf.operators.ConstantOperator`.

    Parameters
    ----------
    training_parameters : list of s scalars or (p,) 1D ndarrays
        Parameter values for which the operator vector is known or
        will be inferred from data. If not provided in the constructor,
        use :meth:`set_training_parameters` later.
    entries : list of s ndarrays, or None
        Operator vectors corresponding to the ``training_parameters``.
        If not provided in the constructor, use :meth:`set_entries` later.
    InterpolatorClass : type or None
        Class for the elementwise interpolation. Must obey the syntax

            >>> interpolator = InterpolatorClass(data_points, data_values)
            >>> interpolator_evaluation = interpolator(new_data_point)

        This can be, e.g., a class from :mod:`scipy.interpolate`.
        If ``None`` (default), use :class:`scipy.interpolate.CubicSpline`
        for one-dimensional parameters and
        :class:`scipy.interpolate.LinearNDInterpolator` otherwise.
        If not provided in the constructor, use :meth:`set_interpolator` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    _OperatorClass = ConstantOperator


class InterpLinearOperator(_InterpOperator):
    r"""Parametric linear operator
    :math:`\Ophat_{\ell}(\qhat,\u;\bfmu) = \Ahat(\bfmu)\qhat`
    where :math:`\Ahat(\bfmu) \in \RR^{r \times r}` and
    the parametric dependence is handled with elementwise interpolation.

    .. math::
       \Ahat(\bfmu) = \textrm{interpolate}(
       (\bfmu_0,\Ahat^{(0)}),\ldots,(\bfmu_{s-1},\Ahat^{(s-1)}); \bfmu)

    Here,

    * :math:`\bfmu_0,\ldots,\bfmu_{s-1}\in\RR^p`
      are the (fixed) training parameter values, and
    * :math:`\Ahat^{(i)} = \Ahat(\bfmu_i) \in \RR^{r \times r}`
      is the operator matrix for training parameter value :math:`\bfmu_i`.

    See :class:`opinf.operators.LinearOperator`

    Parameters
    ----------
    training_parameters : list of s scalars or (p,) 1D ndarrays
        Parameter values for which the operator entries are known
        or will be inferred from data. If not provided in the constructor,
        use :meth:`set_training_parameters` later.
    entries : list of s ndarrays, or None
        Operator entries corresponding to the ``training_parameters``.
        If not provided in the constructor, use :meth:`set_entries` later.
    InterpolatorClass : type or None
        Class for the elementwise interpolation. Must obey the syntax

            >>> interpolator = InterpolatorClass(data_points, data_values)
            >>> interpolator_evaluation = interpolator(new_data_point)

        This can be, e.g., a class from :mod:`scipy.interpolate`.
        If ``None`` (default), use :class:`scipy.interpolate.CubicSpline`
        for one-dimensional parameters and
        :class:`scipy.interpolate.LinearNDInterpolator` otherwise.
        If not provided in the constructor, use :meth:`set_interpolator` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    _OperatorClass = LinearOperator


class InterpQuadraticOperator(_InterpOperator):
    r"""Parametric quadratic operator
    :math:`\Ophat_{\ell}(\qhat,\u;\bfmu) = \Hhat(\bfmu)[\qhat\otimes\qhat]`
    where :math:`\Ahat(\bfmu) \in \RR^{r \times r^2}` and
    the parametric dependence is handled with elementwise interpolation.

    .. math::
       \Hhat(\bfmu) = \textrm{interpolate}(
       (\bfmu_0,\Hhat^{(0)}),\ldots,(\bfmu_{s-1},\Hhat^{(s-1)}); \bfmu)

    Here,

    * :math:`\bfmu_0,\ldots,\bfmu_{s-1}\in\RR^p`
      are the (fixed) training parameter values, and
    * :math:`\Hhat^{(i)} = \Hhat(\bfmu_i) \in \RR^{r \times r^2}`
      are the operator entries evaluated at the training parameter values.

    See :class:`opinf.operators.QuadraticOperator`.

    Parameters
    ----------
    training_parameters : list of s scalars or (p,) 1D ndarrays
        Parameter values for which the operator entries are known
        or will be inferred from data. If not provided in the constructor,
        use :meth:`set_training_parameters` later.
    entries : list of s ndarrays, or None
        Operator entries corresponding to the ``training_parameters``.
        If not provided in the constructor, use :meth:`set_entries` later.
    InterpolatorClass : type or None
        Class for the elementwise interpolation. Must obey the syntax

            >>> interpolator = InterpolatorClass(data_points, data_values)
            >>> interpolator_evaluation = interpolator(new_data_point)

        This can be, e.g., a class from :mod:`scipy.interpolate`.
        If ``None`` (default), use :class:`scipy.interpolate.CubicSpline`
        for one-dimensional parameters and
        :class:`scipy.interpolate.LinearNDInterpolator` otherwise.
        If not provided in the constructor, use :meth:`set_interpolator` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    _OperatorClass = QuadraticOperator


class InterpCubicOperator(_InterpOperator):
    r"""Parametric cubic operator
    :math:`\Ophat_{\ell}(\qhat,\u;\bfmu)
    = \Ghat(\bfmu)[\qhat\otimes\qhat\otimes\qhat]`
    where :math:`\Ghat(\bfmu) \in \RR^{r \times r^3}` and
    the parametric dependence is handled with elementwise interpolation.

    .. math::
       \Ghat(\bfmu) = \textrm{interpolate}(
       (\bfmu_0,\Ghat^{(0)}),\ldots,(\bfmu_{s-1},\Ghat^{(s-1)}); \bfmu)

    Here,

    * :math:`\bfmu_0,\ldots,\bfmu_{s-1}\in\RR^p`
      are the (fixed) training parameter values, and
    * :math:`\Ghat^{(i)} = \Ghat(\bfmu_i) \in \RR^{r \times r^3}`
      are the operator entries evaluated at the training parameter values.

    See :class:`opinf.operators.CubicOperator`.

    Parameters
    ----------
    training_parameters : list of s scalars or (p,) 1D ndarrays
        Parameter values for which the operator entries are known
        or will be inferred from data. If not provided in the constructor,
        use :meth:`set_training_parameters` later.
    entries : list of s ndarrays, or None
        Operator entries corresponding to the ``training_parameters``.
        If not provided in the constructor, use :meth:`set_entries` later.
    InterpolatorClass : type or None
        Class for the elementwise interpolation. Must obey the syntax

            >>> interpolator = InterpolatorClass(data_points, data_values)
            >>> interpolator_evaluation = interpolator(new_data_point)

        This can be, e.g., a class from :mod:`scipy.interpolate`.
        If ``None`` (default), use :class:`scipy.interpolate.CubicSpline`
        for one-dimensional parameters and
        :class:`scipy.interpolate.LinearNDInterpolator` otherwise.
        If not provided in the constructor, use :meth:`set_interpolator` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    _OperatorClass = CubicOperator


class InterpInputOperator(_InterpOperator, InputMixin):
    r"""Parametric input operator
    :math:`\Ophat_{\ell}(\qhat,\u;\bfmu) = \Bhat(\bfmu)\u`
    where :math:`\Bhat(\bfmu) \in \RR^{r \times m}` and
    the parametric dependence is handled with elementwise interpolation.

    .. math::
       \Bhat(\bfmu) = \textrm{interpolate}(
       (\bfmu_0,\Bhat^{(0)}),\ldots,(\bfmu_{s-1},\Bhat^{(s-1)}); \bfmu)

    Here,

    * :math:`\bfmu_0,\ldots,\bfmu_{s-1}\in\RR^p`
      are the (fixed) training parameter values, and
    * :math:`\Bhat^{(i)} = \Bhat(\bfmu_i) \in \RR^{r \times m}`
      are the operator entries evaluated at the training parameter values.


    See :class:`opinf.operators.InputOperator`.

    Parameters
    ----------
    training_parameters : list of s scalars or (p,) 1D ndarrays
        Parameter values for which the operator entries are known
        or will be inferred from data. If not provided in the constructor,
        use :meth:`set_training_parameters` later.
    entries : list of s ndarrays, or None
        Operator entries corresponding to the ``training_parameters``.
        If not provided in the constructor, use :meth:`set_entries` later.
    InterpolatorClass : type or None
        Class for the elementwise interpolation. Must obey the syntax

            >>> interpolator = InterpolatorClass(data_points, data_values)
            >>> interpolator_evaluation = interpolator(new_data_point)

        This can be, e.g., a class from :mod:`scipy.interpolate`.
        If ``None`` (default), use :class:`scipy.interpolate.CubicSpline`
        for one-dimensional parameters and
        :class:`scipy.interpolate.LinearNDInterpolator` otherwise.
        If not provided in the constructor, use :meth:`set_interpolator` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    _OperatorClass = InputOperator

    @property
    def input_dimension(self):
        r"""Dimension of the input :math:`\u` that the operator acts on."""
        return None if self.entries is None else self.shape[1]


class InterpStateInputOperator(_InterpOperator, InputMixin):
    r"""Parametric state-input operator
    :math:`\Ophat_{\ell}(\qhat,\u;\bfmu) = \Nhat(\bfmu)[\u\otimes\qhat]`
    where :math:`\Nhat(\bfmu) \in \RR^{r \times rm}` and
    the parametric dependence is handled with elementwise interpolation.

    .. math::
       \Nhat(\bfmu) = \textrm{interpolate}(
       (\bfmu_0,\Nhat^{(0)}),\ldots,(\bfmu_{s-1},\Nhat^{(s-1)}); \bfmu)

    Here,

    * :math:`\bfmu_0,\ldots,\bfmu_{s-1}\in\RR^p`
      are the (fixed) training parameter values, and
    * :math:`\Nhat^{(i)} = \Nhat(\bfmu_i) \in \RR^{r \times rm}`
      are the operator entries evaluated at the training parameter values.

    See :class:`opinf.operators.StateInputOperator`.

    Parameters
    ----------
    training_parameters : list of s scalars or (p,) 1D ndarrays
        Parameter values for which the operator entries are known
        or will be inferred from data. If not provided in the constructor,
        use :meth:`set_training_parameters` later.
    entries : list of s ndarrays, or None
        Operator entries corresponding to the ``training_parameters``.
        If not provided in the constructor, use :meth:`set_entries` later.
    InterpolatorClass : type or None
        Class for the elementwise interpolation. Must obey the syntax

            >>> interpolator = InterpolatorClass(data_points, data_values)
            >>> interpolator_evaluation = interpolator(new_data_point)

        This can be, e.g., a class from :mod:`scipy.interpolate`.
        If ``None`` (default), use :class:`scipy.interpolate.CubicSpline`
        for one-dimensional parameters and
        :class:`scipy.interpolate.LinearNDInterpolator` otherwise.
        If not provided in the constructor, use :meth:`set_interpolator` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    _OperatorClass = StateInputOperator

    @property
    def input_dimension(self):
        r"""Dimension of the input :math:`\u` that the operator acts on."""
        if self.entries is None:
            return None
        r, rm = self.shape
        return rm // r


# Utilities ===================================================================
def is_interpolated(obj) -> bool:
    """Return ``True`` if ``obj`` is a interpolated operator object."""
    return isinstance(obj, _InterpOperator)


def nonparametric_to_interpolated(OpClass: type) -> type:
    """Get the interpolated operator class corresponding to a nonparametric
    operator class.

    """
    for InterpClassName in __all__:
        InterpClass = eval(InterpClassName)
        if not isinstance(InterpClass, type) or not issubclass(
            InterpClass, _InterpOperator
        ):  # pragma: no cover
            continue
        if InterpClass._OperatorClass is OpClass:
            return InterpClass
    raise TypeError(
        f"_InterpOperator for class '{OpClass.__name__}' not found"
    )


# Deprecations ================================================================
class InterpolatedConstantOperator(InterpConstantOperator):
    def __init__(
        self,
        training_parameters=None,
        entries=None,
        InterpolatorClass: type = None,
        fromblock=False,
    ):
        warnings.warn(
            "InterpolatedConstantOperator has been renamed "
            "and will be removed in an upcoming release, use "
            "InterpConstantOperator",
            DeprecationWarning,
        )
        InterpConstantOperator.__init__(
            self,
            training_parameters=training_parameters,
            entries=entries,
            InterpolatorClass=InterpolatorClass,
            fromblock=fromblock,
        )


class InterpolatedLinearOperator(InterpLinearOperator):
    def __init__(
        self,
        training_parameters=None,
        entries=None,
        InterpolatorClass: type = None,
        fromblock=False,
    ):
        warnings.warn(
            "InterpolatedLinearOperator has been renamed "
            "and will be removed in an upcoming release, use "
            "InterpLinearOperator",
            DeprecationWarning,
        )
        InterpLinearOperator.__init__(
            self,
            training_parameters=training_parameters,
            entries=entries,
            InterpolatorClass=InterpolatorClass,
            fromblock=fromblock,
        )


class InterpolatedQuadraticOperator(InterpQuadraticOperator):
    def __init__(
        self,
        training_parameters=None,
        entries=None,
        InterpolatorClass: type = None,
        fromblock=False,
    ):
        warnings.warn(
            "InterpolatedQuadraticOperator has been renamed "
            "and will be removed in an upcoming release, use "
            "InterpQuadraticOperator",
            DeprecationWarning,
        )
        InterpQuadraticOperator.__init__(
            self,
            training_parameters=training_parameters,
            entries=entries,
            InterpolatorClass=InterpolatorClass,
            fromblock=fromblock,
        )


class InterpolatedCubicOperator(InterpCubicOperator):
    def __init__(
        self,
        training_parameters=None,
        entries=None,
        InterpolatorClass: type = None,
        fromblock=False,
    ):
        warnings.warn(
            "InterpolatedCubicOperator has been renamed "
            "and will be removed in an upcoming release, use "
            "InterpCubicOperator",
            DeprecationWarning,
        )
        InterpCubicOperator.__init__(
            self,
            training_parameters=training_parameters,
            entries=entries,
            InterpolatorClass=InterpolatorClass,
            fromblock=fromblock,
        )


class InterpolatedInputOperator(InterpInputOperator):
    def __init__(
        self,
        training_parameters=None,
        entries=None,
        InterpolatorClass: type = None,
        fromblock=False,
    ):
        warnings.warn(
            "InterpolatedInputOperator has been renamed "
            "and will be removed in an upcoming release, use "
            "InterpInputOperator",
            DeprecationWarning,
        )
        InterpInputOperator.__init__(
            self,
            training_parameters=training_parameters,
            entries=entries,
            InterpolatorClass=InterpolatorClass,
            fromblock=fromblock,
        )


class InterpolatedStateInputOperator(InterpStateInputOperator):
    def __init__(
        self,
        training_parameters=None,
        entries=None,
        InterpolatorClass: type = None,
        fromblock=False,
    ):
        warnings.warn(
            "InterpolatedStateInputOperator has been renamed "
            "and will be removed in an upcoming release, use "
            "InterpStateInputOperator",
            DeprecationWarning,
        )
        InterpStateInputOperator.__init__(
            self,
            training_parameters=training_parameters,
            entries=entries,
            InterpolatorClass=InterpolatorClass,
            fromblock=fromblock,
        )
